
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include "connectivityPage.h"
#include <QtDebug>
#include <QHttp>
#include <QUrl>

ConnectivityPage::ConnectivityPage(Farm *farm, ManageConnectionDialog *parent)
    : ManageConnectionPage(farm, parent)
{
    setupUi(this);

    buttonGroup = new QButtonGroup;
    buttonGroup->addButton(useLocationDefaultCheckBox);
    buttonGroup->addButton(alwaysPromptLoactionCheckBox);
    buttonGroup->setExclusive(true);

    connect(locationComboBox, SIGNAL(currentIndexChanged(int)),
            SLOT(currentLocationChanged(int)));

    connect(renameLocationButton, SIGNAL(clicked()), SLOT(renameLocation()));
    connect(addLocationButton, SIGNAL(clicked()), SLOT(addLocation()));
    connect(removeLocationButton, SIGNAL(clicked()), SLOT(removeLocation()));

    connect(brokersList, SIGNAL(doubleClicked(const QModelIndex &)),
            SLOT(editItem(const QModelIndex &)));
    connect(brokersList, SIGNAL(itemSelectionChanged()),
            SLOT(checkSelectedBroker()));

    connect(addBrokerButton, SIGNAL(clicked()), SLOT(addBroker()));
    connect(removeBrokerButton, SIGNAL(clicked()), SLOT(removeBroker()));
    connect(upBrokerButton, SIGNAL(clicked()), SLOT(upBroker()));
    connect(downBrokerButton, SIGNAL(clicked()), SLOT(downBroker()));
    connect(testBrokerButton, SIGNAL(clicked()), SLOT(testBroker()));
    connect(&http, SIGNAL( done( bool )), this, SLOT( handleTestBroker( bool )));

    checkSelectedBroker();
}

ConnectivityPage::~ConnectivityPage()
{

}

QString ConnectivityPage::showInputDialog(const QString & title, const QString & label, bool * ok, const QString & initText)
{
    QDialog* inputDialog = new QDialog(this);
    inputDialog->setWindowTitle(title);
    inputDialog->setMinimumWidth( 400 );

    QVBoxLayout* mainLayout = new QVBoxLayout();
    QHBoxLayout* frstLine = new QHBoxLayout();
    QHBoxLayout* scndLine = new QHBoxLayout();
    QHBoxLayout* thrdLine = new QHBoxLayout();

    mainLayout->addLayout(frstLine);
    mainLayout->addLayout(scndLine);
    mainLayout->addLayout(thrdLine);

    QLabel* text = new QLabel(label);
    frstLine->addWidget(text);

    QLineEdit* inputLine = new QLineEdit(initText);
    scndLine->addWidget(inputLine);

    QPushButton* okButton = new QPushButton(tr("OK"));
    QPushButton* cancelButton = new QPushButton(tr("Cancel"));
    thrdLine->addWidget(okButton);
    thrdLine->addWidget(cancelButton);

    inputDialog->setLayout(mainLayout);

    connect(okButton, SIGNAL(clicked()), inputDialog, SLOT(accept()));
    connect(cancelButton, SIGNAL(clicked()), inputDialog, SLOT(reject()));

    *ok = (inputDialog->exec() == QDialog::Accepted);
    return inputLine->text();
}

void ConnectivityPage::addBroker()
{
    bool ok;
    QString text;

    do
    {
        text = showInputDialog(tr("Add Connection Broker"), tr("New Connection Broker:"), &ok);

        if (ok)
        {
            text.remove("http://");
            text.remove("https://");

            if (!text.isEmpty() && !text.trimmed().isEmpty())
            {
                if (brokersList->findItems(text, Qt::MatchExactly).isEmpty())
                {
                    brokersList->addItem(text);
                    break;
                }
                else
                {
                    QMessageBox::information(this, "Brokers list",
                            "The list already contains this item");
                    ok = false;
                }
            }
            else
            {
                QMessageBox::information(this, "Brokers list",
                            "Please enter a value or click Cancel");
                ok = false;
            }
        }
        else
        {
            break;
        }
    }
    while(!ok);
}

void ConnectivityPage::removeBroker()
{
    if (QMessageBox::question(this, this->windowTitle(),
                              tr("Are you sure you want to delete the broker [") + brokersList->currentItem()->text() + "]?",
                              QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
    {
        return;
    }

    if (brokersList->currentItem() != NULL)
    	brokersList->takeItem(brokersList->currentRow());
	else
	    QMessageBox::critical(this, "Remove broker",
	    		"There are no brokers selected");
}

void ConnectivityPage::upBroker()
{
	int currentIndex = brokersList->currentRow();
	if (currentIndex > 0)
	{
		QListWidgetItem *tempItem = brokersList->takeItem(currentIndex);
		brokersList->insertItem(currentIndex - 1, tempItem->text());
	}
	brokersList->setFocus();
}

void ConnectivityPage::downBroker()
{
	int currentIndex = brokersList->currentRow();
	if (currentIndex < brokersList->count() - 1)
	{
		QListWidgetItem *tempItem = brokersList->takeItem(currentIndex);
		brokersList->insertItem(currentIndex + 1, tempItem->text());
	}
	brokersList->setFocus();
}


void ConnectivityPage::testBroker()
{
    // Check that something actually is selected

    if ( brokersList->currentIndex().row() < 0 )
    {
        QMessageBox::warning(this, "Test Broker Connectivity",
                             "No Broker is currently selected.  Please select a broker to test");
        return;
    }

    // Check to see if the broker can be contacted
    if (protocolComboBox->currentText() == "HTTP")
    {
        http.setHost( brokersList->currentItem()->text(), portLineEdit->text().toInt());
    }
    else
    {
        http.setHost( brokersList->currentItem()->text(), QHttp::ConnectionModeHttps, portLineEdit->text().toInt());
    }
    http.head(QUrl::toPercentEncoding("/"));
}

void ConnectivityPage::handleTestBroker( bool errors )
{
    if ( errors )
    {
        QMessageBox::warning(this, "Test Broker Connectivity",
                        tr("The selected broker '%2:%3' could not be reached.  The reason reported is: '%1'").
                        arg(http.errorString(), brokersList->currentItem()->text(), portLineEdit->text()));
    }
    else
    {
        QMessageBox::information(this, "Test Broker Connectivity",
                        tr("The selected broker '%1:%2' appears to be active").arg(
                               brokersList->currentItem()->text(), portLineEdit->text()));
    }
}


void ConnectivityPage::editItem(const QModelIndex &)
{
    bool ok;
    QString currentText = brokersList->currentItem()->text();
    QString text = QInputDialog::getText(this, "Broker address",
            "Broker address:", QLineEdit::Normal, currentText, &ok);
    if (ok && !text.isEmpty() && text != currentText)
    {
    	text.remove("http://");
    	text.remove("https://");
        if (brokersList->findItems(text, Qt::MatchExactly).isEmpty())
		brokersList->currentItem()->setText(text);

        else
		QMessageBox::information(this, "Brokers list",
    	    		"The list already contains this address");
    }
}

void ConnectivityPage::setEnabledBrokerButtons(bool enabled)
{
    removeBrokerButton->setEnabled(enabled);
    upBrokerButton->setEnabled(enabled);
    downBrokerButton->setEnabled(enabled);
    testBrokerButton->setEnabled(enabled);
}

void ConnectivityPage::checkSelectedBroker()
{
    QList<QListWidgetItem*> itemsList = brokersList->selectedItems();
    if (itemsList.count() == 0)
    {
        setEnabledBrokerButtons(false);
    }
    else
    {
        setEnabledBrokerButtons(true);
    }
}

void ConnectivityPage::renameLocation()
{
	QString currentText = locationComboBox->currentText();
	if (currentText.isEmpty())
	{
	    return;
	}

	bool ok;

	do
	{
        QString text = QInputDialog::getText(this, "Rename location",
                "Please enter the new name for [" + currentText + "]",
                QLineEdit::Normal, currentText, &ok);

        if (ok && currentText != text)
        {
            if (!text.isEmpty() && !text.trimmed().isEmpty())
            {

                if (locationComboBox->findText(text, Qt::MatchExactly) == -1)
                {
                    locationComboBox->setItemText(locationComboBox->currentIndex(), text);
                    Location *location = currentFarm->getCurrentLocation();
                    if (location != 0)
                    {
                        location->setName(text);
                        break;
                    }
                }
                else
                {
                    QMessageBox::information(this, "Locations list",
                        "The list already contains this location");
                    ok = false;
                }
            }
            else
            {
                QMessageBox::information(this, "Locations list",
                    "Please enter a value or click Cancel");
                ok = false;
            }
        }
        else
        {
            break;
        }
	}
	while(!ok);
}

void ConnectivityPage::addLocation()
{
    // Initialization new location for current farm
    Location *newLocation = currentFarm->createNewLocation();

    // Add new location in current locations list
    int currentIndex = currentFarm->addLocation(newLocation);

    // Updating locations list
    this->updateExistingLocationsList();
    locationComboBox->setCurrentIndex(currentIndex);
}

void ConnectivityPage::removeLocation()
{
    // Get current location index
    int currentIndex = locationComboBox->currentIndex();

    // Remove location at current index
    currentFarm->removeLocation(currentIndex);

    // Updating locations list
    this->updateExistingLocationsList();
    locationComboBox->setCurrentIndex(currentIndex - 1);
}

void ConnectivityPage::saveLocationSettings()
{
	// No current farm
        if (currentFarm == 0)
		return;

        Location *location = currentFarm->getCurrentLocation();
	// No current location
        if (location == 0)
            return;

	// Create brokers list
	QStringList strList;
	for (int i = 0; i < brokersList->count(); ++i)
	{
		strList << brokersList->item(i)->text();
	}

    location->setServers(strList);

    location->setPort(portLineEdit->text());
    location->setProtocol(protocolComboBox->currentText().toLower());

	if (useLocationDefaultCheckBox->isChecked())
	{
        currentFarm->setDefaultLocation(locationComboBox->findText(location->getName()));
	}
}

void ConnectivityPage::loadLocationSettings()
{
        if (currentFarm == 0)
            return;

        int currentLocationIndex = locationComboBox->currentIndex();
        Location *location = currentFarm->getLocation(currentLocationIndex);
	// No current location
        if (location == 0)
		return;

        // Autolaunch location
        if (!currentFarm->isPromptLocation())
        {
            int defaultLocationIndex = currentFarm->getDefaultLocation();
            if (currentLocationIndex == defaultLocationIndex)
            {
                useLocationDefaultCheckBox->setChecked(true);
            }
            else
            {
                // sorry...
                buttonGroup->setExclusive(false);

                useLocationDefaultCheckBox->setChecked(false);

                buttonGroup->setExclusive(true);
            }
        }

	portLineEdit->setText(location->getPort());

	brokersList->clear();
        brokersList->addItems(location->getServers());

	// Find and set protocol in list
	QString protocol = location->getProtocol();
	int protocolIndex = protocolComboBox->findText(protocol, Qt::MatchFixedString);
	if (protocolIndex < 0)
		protocolIndex = 0;
        protocolComboBox->setCurrentIndex(protocolIndex);
}

void ConnectivityPage::updateExistingLocationsList()
{
    // No existing farm
    if (currentFarm == 0)
            return;

    disconnect(locationComboBox, SIGNAL(currentIndexChanged(int)),
            this, SLOT(currentLocationChanged(int)));

    // Creating locations list
    QStringList locationsList = currentFarm->getLocationNamesList();
    locationComboBox->clear();
    locationComboBox->addItems(locationsList);

    // Loading current location
    int currentIndex = currentFarm->getLastLocation();
    locationComboBox->setCurrentIndex(currentIndex);
    this->loadLocationSettings();
    connect(locationComboBox, SIGNAL(currentIndexChanged(int)),
            SLOT(currentLocationChanged(int)));
    currentFarm->setCurrentLocation(currentIndex);
    this->updateLocationName(locationComboBox->currentText());

    // Location buttons management
    bool removeDisable = (locationComboBox->count() == 1);
    removeLocationButton->setEnabled(!removeDisable);
}

void ConnectivityPage::currentLocationChanged(int index)
{
    if (currentFarm == 0)
        return;

    // Break if index is incorrect
    if (index < 0 )
    {
        currentFarm->setCurrentLocation(0);
        return;
    }

    // Save current location settings
//    Location *currentLocation = currentFarm->getCurrentLocation();
//    if (currentLocation != 0 &&
//        currentLocation->getName() != locationComboBox->currentText())
    this->saveLocationSettings();

    // Change current location
    currentFarm->setCurrentLocation(index);

    // Load current location settings
    this->loadLocationSettings();

    // Update current location name on form
    this->updateLocationName(locationComboBox->currentText());
}

void ConnectivityPage::loadFarmSettings()
{
    // Farm doesn`t existing
    if (currentFarm == 0)
        return;

    // What widget was presented
    bool settingsHided = currentFarm->isHideSettings();
    if (settingsHided)
        brokersStackedWidget->setCurrentIndex(1);
    else
        brokersStackedWidget->setCurrentIndex(0);

    renameLocationButton->setEnabled(!settingsHided);
    addLocationButton->setEnabled(!settingsHided);

    // Prompted for location
    bool promptForLocation = currentFarm->isPromptLocation();
    alwaysPromptLoactionCheckBox->setChecked(promptForLocation);

    retryGroupBox->setChecked(currentFarm->waitForDesktop());
    if (currentFarm->waitForDesktop())
    {
        retrySpinBox->setValue(currentFarm->getNumberOfRetries());
        waitSpinBox->setValue(currentFarm->getWaitTime());
    }

    // Update existing locations list
    this->updateExistingLocationsList();
}

void ConnectivityPage::saveFarmSettings()
{
    // Farm doesn`t existing
    if (currentFarm == 0)
        return;

    // Prompted for location
    bool promptForLocation = alwaysPromptLoactionCheckBox->isChecked();
    currentFarm->setPromptLocation(promptForLocation);

    if (currentFarm->getCurrentLocation() != 0)
        this->saveLocationSettings();

    currentFarm->setWaitForDesktop(retryGroupBox->isChecked());
    if (retryGroupBox->isChecked())
    {
        currentFarm->setWaitTime(waitSpinBox->value());
        currentFarm->setNumberOfRetries(retrySpinBox->value());
    }
}

/*virtual*/ void ConnectivityPage::showEvent(QShowEvent *)
{
    if (currentFarm == 0)
        return;

    // Update existing locations list
    this->updateExistingLocationsList();
}

/*virtual*/ void ConnectivityPage::hideEvent(QHideEvent *)
{
    if (currentFarm == 0)
        return;

    int currentIndex = locationComboBox->currentIndex();

    // If no index
    if (currentIndex < 0)
        return;

    currentFarm->setLastLocation(currentIndex);

    if (currentFarm->getCurrentLocation() != 0)
        this->saveLocationSettings();
}

void ConnectivityPage::updateLocationName(const QString & text)
{
	brokerOptionsGroupBox->setTitle(tr("Properties for Location [") + text + "]");
}
