
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include "manageConnectionDialog.h"
#include "manageConnectionPage.h"
#include "TitlePage/titlePage.h"
#include "ConnectivityPage/connectivityPage.h"
#include "FirewallProxyPage/firewallProxyPage.h"
#include "CredentialsPage/credentialsPage.h"
#include "DisplayPage/displayPage.h"
#include "LocalResourcesPage/localResourcesPage.h"
#include "UserExperiencePage/userExperiencePage.h"
#include "PasswordManagementPage/passwordManagementPage.h"
#include "AutoLaunchPage/autoLaunchPage.h"
#include "ConfigurationSourcePage/configurationSourcePage.h"

#include <QtDebug>

ManageConnectionDialog::ManageConnectionDialog(FarmSettings *settings, bool wizardMode,
                           const QString & actuallyFarmName, QWidget *parent/* = 0*/) : QDialog(parent)
{
    setupUi(this);

    // Copying settings
    farmSettings = settings;
    QString farmName = actuallyFarmName;
    if (farmName.isEmpty() && !farmSettings->getFarmNamesList().isEmpty())
    {
        farmName = farmSettings->getFarm(0)->getName();
    }

    connect(okButton, SIGNAL(clicked()), SLOT(accept()));
    connect(cancelButton, SIGNAL(clicked()), SLOT(cancel()));
    connect(backButton, SIGNAL(clicked()), SLOT(back()));
    connect(nextButton, SIGNAL(clicked()), SLOT(next()));

    connect(leftList, SIGNAL(currentRowChanged(int)),
			SLOT(currentPageChanged(int)));

    // Setting wizard or edit mode of dialog
    this->setDialogMode(wizardMode);

    // Set editing farm
    this->setDialogFarm(wizardMode, farmName);
    this->createConnectionDialog();

    // Loading farm settings
    this->loadFarmSettings();
}

ManageConnectionDialog::~ManageConnectionDialog()
{

}

void ManageConnectionDialog::saveFarmSettings()
{
    if (currentFarm == 0)
        return;

    ManageConnectionPage *pageWidget;

	for (int i = 1; i < pagesStackedWidget->count(); i++)
	{
		pageWidget = (ManageConnectionPage *) pagesStackedWidget->widget(i);
		if (pageWidget != NULL)
		{
			pageWidget->saveFarmSettings();
		}
	}
}

void ManageConnectionDialog::loadFarmSettings()
{
    if (currentFarm == 0)
        return;

    ManageConnectionPage *pageWidget;

    for (int i = 1; i < pagesStackedWidget->count(); i++)
    {
        pageWidget = (ManageConnectionPage *) pagesStackedWidget->widget(i);
        if (pageWidget != NULL)
        {
            pageWidget->updateFarm(currentFarm);
            pageWidget->loadFarmSettings();
        }
    }

    QString farmName = currentFarm->getName();
    this->updateFarmName(farmName);
}

bool ManageConnectionDialog::isFarmLocationsCorrect(Farm* farm)
{
    QStringList locList = farm->getLocationNamesList();
    for (int i = 0; i < locList.count(); ++i)
    {
        Location* loc = farm->getLocation(i);
        if (loc->isSsl() && loc->getSslGateway().isEmpty())
        {
            QMessageBox::information(this, this->windowTitle(), tr("Please specify the SSL gateway server from farm [") + farm->getName() + tr("], location [") + loc->getName() + "]");
            leftList->setCurrentRow (3);
            return false;
        }
    }
    return true;
}

void ManageConnectionDialog::accept()
{
    this->saveFarmSettings();

    if (!isFarmLocationsCorrect(currentFarm))
    {
        return;
    }

    // Save farm options in settings file
    if (p_isWizardMode)
    {
        farmSettings->addFarm(currentFarm);
    }
    else
    {
        *farmSettings->getFarm(p_currentFarmName) =
                *currentFarm;
    }

    return QDialog::accept();
}

void ManageConnectionDialog::cancel()
{

	QDialog::rejected();
	this->close();
}

void ManageConnectionDialog::next()
{
	// Additive functions, when next button was pushed
	ManageConnectionPage *currentPage = (ManageConnectionPage *)pagesStackedWidget->currentWidget();
	bool bResult = currentPage->nextEvents();
	if (!bResult)
		return;
	int currentIndex = leftList->currentRow();
	int count = leftList->count();
	if (currentIndex == count - 2)
	{
		okButton->setText("&Finish");
		okButton->show();
		nextButton->hide();
	}

	if (currentIndex == 0)
	{
		backButton->setEnabled(true);
	}
	leftList->setCurrentRow(++currentIndex);
}

void ManageConnectionDialog::back()
{
	// Additive functions, when back button was pushed
	ManageConnectionPage *currentPage = (ManageConnectionPage *)pagesStackedWidget->currentWidget();
	currentPage->backEvents();
	int currentIndex = leftList->currentRow();
	int count = leftList->count();
	if (currentIndex == count - 1)
	{
		okButton->setText("&OK");
		okButton->hide();
		nextButton->show();
	}
	if (currentIndex == 1)
	{
	    backButton->setEnabled(false);
	}
	leftList->setCurrentRow(--currentIndex);
}

void ManageConnectionDialog::setDialogMode(bool bWizardMode)
{
    p_isWizardMode = bWizardMode;

    nextButton->setVisible(bWizardMode);
    backButton->setVisible(bWizardMode);

    okButton->setVisible(!bWizardMode);
    leftList->setEnabled(!bWizardMode);
    backButton->setEnabled(!bWizardMode);
}

void ManageConnectionDialog::addPage(QWidget *page)
{
	pagesStackedWidget->addWidget(page);
	leftList->addItem(page->windowTitle());
}

void ManageConnectionDialog::currentPageChanged(int index)
{
	pagesStackedWidget->setCurrentIndex(index+1);
	QWidget *currentWidget = pagesStackedWidget->currentWidget();
	pageTitleLabel->setText("<b>" + currentWidget->windowTitle() + "</b>");
	pageSubTitleLabel->setText(currentWidget->whatsThis());
}

void ManageConnectionDialog::updateFarmName(QString name)
{
    p_currentFarmName = name;

    const int maxNameLength = 100;
    const QString dots = "...";

    if (name.length() > maxNameLength)
    {
        name.resize(maxNameLength);
        name.push_back(dots);
    }
    nameLabel->setText("<b>" + name + "</b>");
}

void ManageConnectionDialog::createConnectionDialog()
{

    // Create title page in dialog
    TitlePage *titlePage = new TitlePage(currentFarm, this);
    this->addPage(titlePage);
    connect(titlePage, SIGNAL(farmSettingsChanged()), SLOT(loadFarmSettings()));

    this->addPage(new ConfigurationSourcePage(currentFarm, this));
    this->addPage(new ConnectivityPage(currentFarm, this));
    this->addPage(new FirewallProxyPage(currentFarm, this));

    bool passThrough = farmSettings->isPassThroughEnabled();
    this->addPage(new CredentialsPage(currentFarm, passThrough, this));

    this->addPage(new DisplayPage(currentFarm, this));
    this->addPage(new LocalResourcesPage(currentFarm, this));
    this->addPage(new UserExperiencePage(currentFarm, this));
    this->addPage(new PasswordManagementPage(currentFarm, this));
    this->addPage(new AutoLaunchPage(currentFarm, this));

    leftList->setCurrentRow(0);
}

void ManageConnectionDialog::setDialogFarm(bool wizardMode, const QString & name)
{
    // Initialize temp working farm
    p_currentFarmName = name;

    // Wizard mode
    if (wizardMode)
    {
        currentFarm = farmSettings->createNewFarm();
    }
    // Edit mode
    else
    {
        currentFarm = new Farm(farmSettings);
        *currentFarm = *farmSettings->getFarm(name);

        int locationsCount = currentFarm->getLocationNamesList().count();

        for(int i = 0; i < locationsCount; ++i)
        {
            Location *loc = new Location();
            *loc = *currentFarm->getLocation(0);
            currentFarm->addLocation(loc);
            currentFarm->removeLocation(0);
        }

        //qDebug() << "Locations list: " << currentFarm->getLocationNamesList();
    }
}

