
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include <QtGui>
#include <QtDebug>
#include "mainWindow.h"

#include "ManageConnectionDialog/manageConnectionDialog.h"

#define APPLICATION_NAME tr("vWorkspace AppPortal")

MainWindow::MainWindow(QWidget *parent)
	: QMainWindow(parent),
	  settings("Quest", APPLICATION_NAME)
{
    this->setWindowTitle(APPLICATION_NAME);
	QWidget * centralWidget = new QWidget;
    mainLayout = new QVBoxLayout;

    treeMenuWidget = new TreeMenuWidget();

	// Initialize farm connection
	currentConnection = new FarmConnection(this);
    connect(currentConnection, SIGNAL(finished(FarmConnection::FinishCode)),
            SLOT(connectionFinished(FarmConnection::FinishCode)));
    connect(currentConnection, SIGNAL(error()), SLOT(connectionError()));
    connect(currentConnection, SIGNAL(startApplication(const QString &)),
            SLOT(runApplication(const QString &)));

    // Progress dialog
    progressDialog = new ProgressDialog(this);
    progressDialog->setWindowTitle(tr("Connection..."));
    progressDialog->setWindowModality(Qt::WindowModal);
    connect(currentConnection, SIGNAL(stateChanged(const QString &)),
            progressDialog, SLOT(setStatusText(const QString &)));
    connect(progressDialog, SIGNAL(canceled()),
            currentConnection, SLOT(reject()));
    connect(currentConnection, SIGNAL(started()), progressDialog, SLOT(start()));

    createActions();
	createMenus();
	createToolBars();
	createVerticalMenus();
	createStatusBar();

	setIcons();

    centralWidget->setLayout(mainLayout);
    setCentralWidget(centralWidget);

	connect(treeMenuWidget, SIGNAL(executeApplication(const QString &)),
            SLOT(runApplication(const QString &)));

    this->readSettings();
}

void MainWindow::createActions()
{
    manageConnectionAction = new QAction (QIcon(":/img/edit_farm_icon.png"),
        tr("Manage Connections..."), this);
    connect(manageConnectionAction, SIGNAL(triggered()), SLOT(editFarm()));

    changeCurrentConnectionAction = new QAction (QIcon(":/img/location_icon.png"),
        tr("Change Current Location..."), this);
    connect(changeCurrentConnectionAction, SIGNAL(triggered()), SLOT(changeCurrentConnection()));
    changeCurrentConnectionAction->setEnabled(false);

    logonAsADifferentUserAction = new QAction(QIcon(":/img/change_user_icon.png"),
        tr("Logon as a Different User..."), this);
    connect(logonAsADifferentUserAction, SIGNAL(triggered()), SLOT(logonAsADifferentUser()));
    logonAsADifferentUserAction->setEnabled(false);

    changePasswordAction = new QAction (QIcon(":/img/password_icon.png"), tr("Change Password..."), this);
    connect(changePasswordAction, SIGNAL(triggered()), this, SLOT(changePassword()));
    changePasswordAction->setEnabled(true);

    refreshApplicationSetAction = new QAction (QIcon(":/img/refresh_icon.png"),
        tr ("Refresh Application Set"), this);
    connect(refreshApplicationSetAction, SIGNAL(triggered()), this, SLOT(refreshApplicationSet()));
    refreshApplicationSetAction->setEnabled(false);

    closeAction = new QAction(QIcon(":/img/exit_icon.png"), tr("Close"), this);
    connect (closeAction, SIGNAL(triggered()), qApp, SLOT(closeAllWindows()));

    // Show or hide menubar
    menuBarAction = new QAction (tr ("MenuBar"), this);
    menuBarAction->setCheckable(true);
    connect(menuBarAction, SIGNAL(triggered()), this, SLOT(menuBarSlot()));

    // Show or hide toolbar
    toolBarAction = new QAction(tr ("ToolBar"), this);
    toolBarAction->setCheckable(true);
    connect(toolBarAction, SIGNAL(triggered()), this, SLOT(toolBarSlot()));

    enableCredentialsPassThroughAction = new QAction(tr("Enable Credentials Pass-Through"), this);
    enableCredentialsPassThroughAction->setCheckable(true);
    enableCredentialsPassThroughAction->setEnabled(false);
    connect(enableCredentialsPassThroughAction, SIGNAL(triggered()), SLOT(enableCredentialsPassThrough()));

    helpAction = new QAction(tr("Help"), this);
    helpAction->setShortcut(tr("F1"));
    helpAction->setEnabled(false);

    // Show AboutBox dialog
    aboutAction = new QAction(tr("About..."), this);
    connect(aboutAction, SIGNAL(triggered()), this, SLOT(about()));
}

void MainWindow::createMenus()
{
    actionMenu = menuBar()->addMenu(tr("&Action"));
    actionMenu->addAction(manageConnectionAction);
    actionMenu->addAction(changeCurrentConnectionAction);
    actionMenu->addAction(logonAsADifferentUserAction);
    actionMenu->addAction(changePasswordAction);
    actionMenu->addSeparator();
    actionMenu->addAction(refreshApplicationSetAction);
    actionMenu->addSeparator();
    actionMenu->addAction(closeAction);

    QMenu *authenticationSubMenuAction = new QMenu (tr ("Authentication"), actionMenu);

    settingsMenu = menuBar()->addMenu (tr ("Settings"));
    settingsMenu->addAction(menuBarAction);
    settingsMenu->addAction(toolBarAction);
    settingsMenu->addSeparator();
    settingsMenu->addMenu(authenticationSubMenuAction);
    authenticationSubMenuAction->addAction(enableCredentialsPassThroughAction);

    helpMenu = menuBar()->addMenu(tr("Help"));
    helpMenu->addAction(helpAction);
    helpMenu->addAction(aboutAction);
}

void MainWindow::createToolBars()
{
    mainToolBar = addToolBar(tr("Action"));

    mainToolBar->addAction(closeAction);
    mainToolBar->addAction(manageConnectionAction);
    mainToolBar->addAction(changeCurrentConnectionAction);
    mainToolBar->addAction(logonAsADifferentUserAction);
    mainToolBar->addAction(changePasswordAction);
    mainToolBar->addAction(refreshApplicationSetAction);

    mainToolBar->setMovable(false);
}

void MainWindow::createVerticalMenus()
{
	farmVerticalMenu = new VerticalMenu(tr("Farms"));
	this->createFarmMenu();
    farmVerticalMenu->addMenuLayout(farmMenuGridLayout);

	actionVerticalMenu = new VerticalMenu(tr("Actions"));
	actionVerticalMenu->addMenuAction(manageConnectionAction);
	actionVerticalMenu->addMenuAction(changeCurrentConnectionAction);
	actionVerticalMenu->addMenuAction(logonAsADifferentUserAction);
	actionVerticalMenu->addMenuAction(changePasswordAction);
	actionVerticalMenu->addMenuAction(refreshApplicationSetAction);
	actionVerticalMenu->addSpacer();

	QGroupBox *groupBox = new QGroupBox;
	groupBox->setTitle(tr("Welcome"));
	groupBox->setFlat(true);

	QLabel *lbl = new QLabel;
        lbl->setText(tr("Select a farm below to load your shortcuts. Once connected, "
                        "click on a folder bar to show the applications."));
	lbl->setWordWrap(true);

    QVBoxLayout *vLayout = new QVBoxLayout;
	vLayout->addWidget(lbl);
    vLayout->addWidget(farmVerticalMenu);
    vLayout->addWidget(actionVerticalMenu, 0, Qt::AlignTop);

    VerticalMenu *appMenu = new VerticalMenu(tr("Applications"));
    appMenu->addWidget(treeMenuWidget);
    vLayout->addWidget(appMenu);
    groupBox->setLayout(vLayout);

    mainLayout->addWidget(groupBox);
}

void MainWindow::createStatusBar()
{
    statusBar()->showMessage(tr("Ready"));
}

void MainWindow::setIcons()
{
    QIcon icon(":/img/provision_icon.png");
	this->setWindowIcon(icon);
}

void MainWindow::toolBarSlot()
{
    mainToolBar->setVisible(toolBarAction->isChecked());
}

void MainWindow::menuBarSlot()
{
    menuBar()->setVisible(menuBarAction->isChecked());
}

void MainWindow::enableCredentialsPassThrough()
{
    bool enablePassThrough = enableCredentialsPassThroughAction->isChecked();
    settings.setPassThroughEnable(enablePassThrough);
}

void MainWindow::about()
{
	AboutDialog *aboutDialog = new AboutDialog("Version 7.0  - Community Edition (This is not supported by Quest Support)", this);
	aboutDialog->exec();
}

/*virtual*/ void MainWindow::contextMenuEvent(QContextMenuEvent *pe)
{
	settingsMenu->exec(pe->globalPos());
}

/*virtual*/ void MainWindow::closeEvent(QCloseEvent *event)
{
    if (QMessageBox::question(this, this->windowTitle(),
                          tr("Are you sure you want to exit Linux AppPortal?"),
                          QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes)
    {
        this->writeSettings();
        event->accept();
    }
    else
    {
        event->ignore();
    }


	// Destroy connection
	//delete currentConnection;
}

void MainWindow::writeSettings()
{
	settings.setValue("ToolBarVisible", mainToolBar->isVisible());
	settings.setValue("MenuBarVisible", menuBar()->isVisible());
    settings.saveFarms();
}

void MainWindow::readSettings()
{
	bool bToolBarVisible = settings.value("ToolBarVisible", false).toBool();
	bool bMenuBarVisible = settings.value("MenuBarVisible", true).toBool();

	toolBarAction->setChecked(bToolBarVisible);
	this->toolBarSlot();
	menuBarAction->setChecked(bMenuBarVisible);
	this->menuBarSlot();

    settings.loadFarms();
    this->updateFarmsList();

    for(int i = 0; i < farmListWidget->count(); ++i)
    {
        QListWidgetItem *item = farmListWidget->item(i);
        QString name = item->text();
        Farm *farm = settings.getFarm(name);
        if (farm->autoConnected())
        {
            currentConnection->setFarm(farm);
            currentConnection->connectTo();
        }
    }

}

void MainWindow::refreshApplicationSet()
{
    QString processName = tr("Refreshing an application list...");
    this->setStatus(processName);

    currentConnection->refreshApplicationsList();
}

void MainWindow::changeCurrentConnection()
{
    QString processName = tr("Changing location...");
    this->setStatus(processName);

    currentConnection->changeCurrentLocation();
}

void MainWindow::logonAsADifferentUser()
{
    QString processName = tr("Logging as a differnet user...");
    this->setStatus(processName);

    currentConnection->logonAsADifferentUser();
}

void MainWindow::connectionFinishedAlt(FarmConnection::FinishCode exitCode, FarmConnection *connection)
{
    // Updating status bar message
    statusBar()->showMessage(tr("Ready"));
    progressDialog->hide();

    switch (exitCode)
    {
    case FarmConnection::Success:
        QMessageBox::information(this, tr("Password Management..."),
                              tr("Password was changed successfully."));
        break;
    case FarmConnection::Error:
        // Displaying last connection error
        QMessageBox::critical(this, tr("Connection error..."),
                              connection->lastError());
        break;
    case FarmConnection::Reject:
        break;
    case FarmConnection::Message:
        // Need to show the message
        QString message = connection->lastError();
        QMessageBox::information(this, tr("Changing password"), message);
        break;
    }
}

void MainWindow::changePassword()
{
    QString processName = tr("Changing password...");
    this->setStatus(processName);
    FarmConnection* connection = new FarmConnection(this);
    connect(connection, SIGNAL(changePasswordFinished(FarmConnection::FinishCode, FarmConnection*)),
                          SLOT(connectionFinishedAlt(FarmConnection::FinishCode, FarmConnection*)));
    connect(connection, SIGNAL(error()), SLOT(connectionError()));

    connect(connection, SIGNAL(stateChanged(const QString &)),
                progressDialog, SLOT(setStatusText(const QString &)));
    connect(progressDialog, SIGNAL(canceled()),
                connection, SLOT(reject()));
    connect(connection, SIGNAL(started()), progressDialog, SLOT(start()));

    connection->changePassword(&settings);
}

void MainWindow::connectionFinished(FarmConnection::FinishCode exitCode)
{
    // Updating status bar message
    statusBar()->showMessage(tr("Ready"));
    progressDialog->hide();

    switch (exitCode)
    {
    case FarmConnection::Success:
        // Connection was succesfully
        treeMenuWidget->updateApplicationsView(currentConnection->applications());

        // Enabling options for our connection
        refreshApplicationSetAction->setEnabled(true);
        changeCurrentConnectionAction->setEnabled(true);
        logonAsADifferentUserAction->setEnabled(true);

        break;
    case FarmConnection::Error:
        // Connection was failed

        // don't do this on launching an application
        if (!currentConnection->isConnected())
        {
            treeMenuWidget->clear();

            // Disabling options for our connection
            refreshApplicationSetAction->setEnabled(false);
            changeCurrentConnectionAction->setEnabled(false);
            logonAsADifferentUserAction->setEnabled(false);
        }

        // Displaying last connection error
        QMessageBox::critical(this, tr("Connection error..."),
                              currentConnection->lastError());
        break;
    case FarmConnection::Reject:
        // Connection was rejected

        break;
    case FarmConnection::Message:
        // Need to show the message
        QString message = currentConnection->lastError();
        QMessageBox::information(this, tr("Changing password"), message);
        break;
    }
}

void MainWindow::connectionError()
{
        // Updating status bar message
        //statusBar()->showMessage(tr("Connection failed"));

        // Displaying last connection error
        //QString lastError = currentConnection->lastError();
        //QMessageBox::critical(this, tr("Connection error..."), lastError);

        // Clearing app list
        //treeMenuWidget->setVisible(false);
        //treeMenuWidget->clear();

	// Disabling options for our connection
        //refreshApplicationSetAction->setEnabled(false);
        //changeCurrentConnectionAction->setEnabled(false);
        //logonAsADifferentUserAction->setEnabled(false);
}

void MainWindow::runApplication(const QString & appName)
{
    Farm * currentFarm = currentConnection->farm();
    if (currentFarm == 0)
        return;

    ApplicationList appList = currentFarm->applications();
    if (appList.isEmpty())
        return;

    Application * app = appList.application(appName);
    if (app == 0)
        return;

    QPixmap * appPixmap = app->largeIcon();
    if (appPixmap != 0)
    {
        progressDialog->setPixmap(*appPixmap);
    }
    progressDialog->setProcessName(tr("Running application: ") + "<i>" +
                                   appName + "</i>");
    currentConnection->executeSelectedApplication(appName);
}

void MainWindow::showStatusMessage(const QString & status)
{
    statusBar()->showMessage(status);
}

void MainWindow::createFarmMenu()
{
    // Create widgets
    farmListWidget = new QListWidget();

    connect(farmListWidget, SIGNAL(itemActivated(QListWidgetItem *)),
            SLOT(runFarm(QListWidgetItem *)));

    // Layout this widgets
    farmMenuGridLayout = new QVBoxLayout();
    farmMenuGridLayout->addWidget(farmListWidget);

    // Updating this widget
    this->updateFarmsList();
}

void MainWindow::updateFarmsList()
{
    farmListWidget->clear();

    QStringList farmNames = settings.getFarmNamesList();

    // Enable work buttons
    bool isFarms = !farmNames.isEmpty();
    farmListWidget->setEnabled(isFarms);

    Farm *farm = currentConnection != NULL ? currentConnection->farm() : NULL;
    bool isConnectedFarm = false;

    foreach (QString name, farmNames)
    {
        QListWidgetItem *widgetItem = NULL;
        if (farm != NULL && name == farm->getName() && farm == settings.getFarm(name))
        {
            isConnectedFarm = true;
            widgetItem = new QListWidgetItem(QIcon(":/img/farm_connected.png"), name);
        }
        else
        {
            widgetItem = new QListWidgetItem(QIcon(":/img/farm_connection.png"), name);
        }

        farmListWidget->addItem(widgetItem);
    }

    if (!isConnectedFarm)
    {
        if (currentConnection != NULL)
        {
            this->currentConnection->disconnectTo();
        }
        this->treeMenuWidget->clear();
        refreshApplicationSetAction->setEnabled(false);
        changeCurrentConnectionAction->setEnabled(false);
        logonAsADifferentUserAction->setEnabled(false);
    }
}

void MainWindow::editFarm()
{
    ManageConnectionDialog *connectionDialog;
    // No farms
    if (farmListWidget->count() == 0)
    {
         connectionDialog = new ManageConnectionDialog(&settings,
                                                            true,
                                                            QString(),
                                                            this);
    }
    else if (farmListWidget->currentItem() == 0)
    {
        connectionDialog = new ManageConnectionDialog(&settings,
                                                            false,
                                                            QString(),
                                                            this);
    }
    else
    {
        QString name = farmListWidget->currentItem()->text();

        // Create edit farm dialog
        connectionDialog = new ManageConnectionDialog(&settings,
                                                            false,
                                                            name,
                                                            this);
    }

    // Farm was changed
    if (connectionDialog->exec() == QDialog::Accepted && currentConnection != NULL)
    {
        this->updateFarmsList();
        Farm *currentFarm = currentConnection->farm();
        if (currentFarm != NULL && currentFarm->needsReconnecting())
        {
            int row = settings.getFarmNamesList().indexOf(currentFarm->getName());
            farmListWidget->setCurrentRow(row);
            runFarm();
            currentFarm->setNeedsReconnecting(false);
        }
    }
    this->updateFarmsList();
}

void MainWindow::runFarm(QListWidgetItem *item)
{
    // Farm name to connect
    QListWidgetItem *farmItem = farmListWidget->currentItem();
    if (farmItem == 0)
        return;

    QString currentFarmName = farmItem->text();

    // Initialize farm connection
    Farm *currentFarm = settings.getFarm(currentFarmName);
    currentConnection->setFarm(currentFarm);

    this->treeMenuWidget->clear();

    QString processName = tr("Connecting to farm...");
    this->setStatus(processName);
    bool result  = false;

    do
    {
        result = currentConnection->connectTo();
        if (currentConnection->lastError().contains(tr("Logon failure")))
        {
            currentConnection->setRequestCredentials(true);
        }

    } while(currentConnection->lastError().contains(tr("Logon failure")));

    // Some troubles
    if (!result)
    {
        QString lastError = currentConnection->lastError();
        if (!lastError.isEmpty())
        {
            QMessageBox::information(this,
                                     tr("Connection error..."),
                                     lastError);
        }
    }
    this->updateFarmsList();
}

void MainWindow::setStatus(const QString & status)
{
    statusBar()->showMessage(status);
    progressDialog->setProcessName(status);

    progressDialog->setPixmap(QPixmap(":/img/earth_icon.png"));
}

void MainWindow::autoConnectedFarmChanged(int state)
{
    // set the AutoConnect option for the current farm and clear for all the others
    farmListWidget->setFocus();
    QListWidgetItem *currentItem = farmListWidget->currentItem();
    if (currentItem == 0)
        return;

    for(int i = 0; i < farmListWidget->count(); ++i)
    {
        QListWidgetItem *item = farmListWidget->item(i);
        QString name = item->text();
        Farm *farm = settings.getFarm(name);
        if (item == currentItem)
            farm->setAutoConnect(state != 0);
        else if (state != 0)
            farm->setAutoConnect(false);
    }
}
