
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


/*
 *  tusbd/fifo.c
 *
 *  Copyright (C) 2007-2008 IncentivesPro
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#include "usbd.h"

struct fifo_descriptor *usbd_fifo_create(void)
{
	struct fifo_descriptor *pfifodesc;

   	pfifodesc = kmalloc(sizeof(*pfifodesc), GFP_KERNEL);

   	if( pfifodesc)
   	{
	  	memset(pfifodesc, 0, sizeof(*pfifodesc));

  		INIT_LIST_HEAD(&pfifodesc->entrylist);
		spin_lock_init(&pfifodesc->lock);
		init_waitqueue_head(&pfifodesc->waitqueue);

		return pfifodesc;
 	}

	return NULL;
}

void usbd_fifo_destroy(struct fifo_descriptor *pfifodesc)
{
	if( !pfifodesc ) return;

	usbd_fifo_clear(pfifodesc);

	kfree(pfifodesc);
}

void usbd_fifo_clear(struct fifo_descriptor *pfifodesc)
{
	spin_lock(&pfifodesc->lock);
	while( !list_empty(&pfifodesc->entrylist) )
	{
		struct list_head *entry = pfifodesc->entrylist.next;
		list_del(entry);
		kfree(entry);
	}
	pfifodesc->num_entries = 0;

	spin_unlock(&pfifodesc->lock);
}

int usbd_fifo_get(struct fifo_descriptor *pfifodesc, void *buf, int bufsize, int user_mode_buffer)
{
	int result = 0;

	if( !pfifodesc ) return -EINVAL;
	if( !buf ) return -EINVAL;
	if( !bufsize ) return -EINVAL;

	spin_lock(&pfifodesc->lock);

	if( !list_empty(&pfifodesc->entrylist) )
	{
		struct fifo_entry	*pfifoentry;

		pfifoentry = (struct fifo_entry*)pfifodesc->entrylist.next;

		// calc the nuber of bytes to copy
		result = MIN_OF(bufsize, pfifoentry->size - pfifoentry->read);

		if( user_mode_buffer )
		{
			if( copy_to_user(buf, pfifoentry->data+pfifoentry->read, result) != 0 )
			{
				TRACE("usbd_fifo_get: cannot copy to user buffer\n");
			}
		} else
		{
			memcpy(buf, pfifoentry->data+pfifoentry->read, result);
		}

		pfifoentry->read += result;

		// is all event copied?
		if( pfifoentry->read == pfifoentry->size )
		{
			list_del(&pfifoentry->entry);
			kfree(pfifoentry);
			pfifodesc->num_entries--;
		}
	}

	spin_unlock(&pfifodesc->lock);

	return result;
}

int usbd_fifo_add(struct fifo_descriptor *pfifodesc, const void *buf, int bufsize, int user_mode_buffer)
{
	struct fifo_entry *pfifoentry;

	if( !pfifodesc ) return -EINVAL;
	if( !buf ) return -EINVAL;

	if( !bufsize ) return -EINVAL;

	if( in_atomic() )
	{
	   	pfifoentry = kmalloc(sizeof(*pfifoentry)+bufsize-1, GFP_ATOMIC); /*TODO: this is wrong!*/
	} else
	{
	   	pfifoentry = kmalloc(sizeof(*pfifoentry)+bufsize-1, GFP_KERNEL);
	}

   	if( pfifoentry )
   	{
		if( user_mode_buffer )
		{
			if( copy_from_user(pfifoentry->data, buf, bufsize) != 0 )
			{
				TRACE("usbd_fifo_get: cannot copy to user buffer\n");
			}
		} else
		{
			memcpy(pfifoentry->data, buf, bufsize);
		}

   		pfifoentry->read = 0;
   		pfifoentry->size = bufsize;

		spin_lock(&pfifodesc->lock);
		list_add_tail(&pfifoentry->entry, &pfifodesc->entrylist);
		pfifodesc->num_entries++;
		spin_unlock(&pfifodesc->lock);

		wake_up(&pfifodesc->waitqueue);

		return 0;
   	}

   	return -ENOMEM;
}

int usbd_fifo_add_pointer(struct fifo_descriptor *pfifodesc, const void *buf, int bufsize, int user_mode_buffer)
{
	struct fifo_entry *pfifoentry;

	if( !pfifodesc ) return -EINVAL;
	if( !buf ) return -EINVAL;

	if( !bufsize ) return -EINVAL;

	if( in_atomic() )
	{
	   	pfifoentry = kmalloc(sizeof(*pfifoentry)+bufsize-1, GFP_ATOMIC); /*TODO: this is wrong!*/
	} else
	{
	   	pfifoentry = kmalloc(sizeof(*pfifoentry)+bufsize-1, GFP_KERNEL);
	}

   	if( pfifoentry )
   	{
		if( user_mode_buffer )
		{
			if( copy_from_user(pfifoentry->data, buf, bufsize) != 0 )
			{
				TRACE("usbd_fifo_get: cannot copy to user buffer\n");
			}
		} else
		{
			memcpy(pfifoentry->data, buf, bufsize);
		}

   		pfifoentry->read = 0;
   		pfifoentry->size = bufsize;

		spin_lock(&pfifodesc->lock);
		list_add_tail(&pfifoentry->entry, &pfifodesc->entrylist);
		pfifodesc->num_entries++;
		spin_unlock(&pfifodesc->lock);

		wake_up(&pfifodesc->waitqueue);

		return 0;
   	}

   	return -ENOMEM;
}
