
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include <QtGui>
#include <QDir>
#include <syslog.h>

#include "mainwindow.h"
#include "conndialog.h"
#include "restartProgressDialog.h"
#include "../common/crypto.h"
#include "../common/errors.h"
#include "../common/code.h"

extern QPushButton *createButton(const QString& text, int width = 0, int height = 0);

QString getPntscErrorString(int errcode)
{
    switch (errcode){
    case NO_ERROR:
        return QString("Operation succeeded");
    case INTERNAL_ERROR:
        return QString("Internal program error");
    case INPUT_PARAMETERS_ERROR:
        return QString("Incorrect input parameters");
    case BROKER_ERROR:
        return QString("Broker error");
    case SYSTEM_ERROR:
        return QString("System error");
    case AUTHENTICATION_ERROR:
        return QString("Authentication error");
    case UNKNOWN_HOSTNAME_ERROR:
        return QString("Unknown hostname error");
    default:
        return QString("Unknown error");
    }
}

//----------------------------------------------------------------------------
ConnectionManagerMainWindow::ConnectionManagerMainWindow()
    : pntsc(NULL)
{
    QWidget *widget = new QWidget;

    // create or open the config file
    // QMessageBox::information(this, "Connection Manager", QDir::homePath() + CONFIG_FILE_NAME);

    QString path = QDir::homePath() + QDir::separator() + XML_CONFIG_PATH;
    QDir dir (path);
    if (!dir.exists())
    {
	QDir().mkpath(path);

	// try to read the config file from the old location
	path = QDir::homePath();
    }
    QFile confFile(path + QDir::separator() + XML_CONFIG_FILE_NAME);
    if (!confFile.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        QMessageBox::critical(this, APPLICATION_NAME, "Can't open the config file.");
        close();
    }

    QXmlStreamReader xml(&confFile);
    while (!xml.atEnd())
    {
        xml.readNext();
        if (xml.isStartElement())
        {
            if ( xml.name() == "Domains" )
                this->readDomains(&xml);
            else if (xml.name() == "AutoLaunchedApps")
                this->readAutoLaunchedApps(&xml);
            else if (xml.name() == "WorkingDirectory")
            {
                workingDirName = QString(xml.readElementText());
                if (workingDirName == "")
                    workingDirName = QDir::homePath();
            }
        }
    }
    if (xml.hasError())
    {
        QMessageBox::warning(this, APPLICATION_NAME, "Reading config file - " + xml.errorString());
    }
/*
    char workingDirStr[1024];
    char autoLaunchStr[1024];
    qint64 len1 = confFile.readLine(workingDirStr, 1024);
    qint64 len2 = confFile.readLine(autoLaunchStr, 1024);
    if (len2 > 0)
    {
        autoLaunchConnectionName = QString(autoLaunchStr).replace(QString("\n"), QString(""));
    }
    if (len1 <= 0 || !QDir(QString(workingDirStr).replace(QString("\n"), QString(""))).exists())
    {
        workingDirName = QDir::homePath();
        confFile.resize(0);
        QTextStream stream(&confFile);
        stream << workingDirName << endl;
        if (autoLaunchConnectionName != "")
        {
            stream << autoLaunchConnectionName << endl;
        }
    }
    else
    {
        workingDirName = QString(workingDirStr).replace(QString("\n"), QString(""));
    }
*/
    confFile.close();

    if (!QDir::setCurrent(QString(workingDirName)) && !QString(workingDirName).isEmpty())
    {
        QMessageBox::critical(this, APPLICATION_NAME,
                              "Can't change directory to" + QString(workingDirName));
        QDir::setCurrent(QDir::homePath());
    }
    label = new QLabel("Connection List (" + QDir::currentPath() + ")");

    connList = new QListWidget;
    connList->setStatusTip("Select a connection");
    autoLaunchCheckBox = new QCheckBox("Automatically launch this connection on startup");
    autoLaunchCheckBox->setChecked(false);

    connect(connList, SIGNAL(doubleClicked ( const QModelIndex &)),
            this, SLOT(editConnection()));
    connect(connList, SIGNAL(currentItemChanged ( QListWidgetItem *, QListWidgetItem *)),
            this, SLOT(checkAutoLaunch(QListWidgetItem *, QListWidgetItem *)));
    connect(autoLaunchCheckBox, SIGNAL(clicked(bool)),
            this, SLOT(setAutoLaunchConnection(bool)));

    createActions();
    createMenus();
    createToolBars();

    // By default disable the connection action buttons
    disableConnectionActions();

    QString fileName ("*.rdp");
    QStringList files = QDir::current().entryList(QStringList(fileName), QDir::Files | QDir::NoSymLinks);

    // If there is more than one connection file, them enable the Start, Edit, Delete actions.
    if ( files.size() > 0 ) enableConnectionActions();

    int i = 0;
    for (; i < files.size(); i++)
    {
        files[i].truncate(files[i].length() - 4);
        connList->addItem(files[i]);
    }

    statusBar()->showMessage(tr("Ready"));

    QVBoxLayout *layout = new QVBoxLayout();
    layout->addWidget(label);
    layout->addWidget(connList);
    layout->addWidget(autoLaunchCheckBox);
    layout->setContentsMargins(0, 0, 0, 10);
    widget->setLayout(layout);

    setCentralWidget(widget);

    setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
    setMinimumSize(QSize(700,450));
    setMaximumSize(QSize(700,450));
    QRect rect = QApplication::desktop()->screenGeometry(this);
    move((rect.width() - width())/2, (rect.height()-height())/2);

    // auto-launch a connection
    if (autoLaunchConnectionName != "")
    {
        connList->setCurrentItem(connList->findItems(autoLaunchConnectionName, Qt::MatchExactly).at(0));
        startConnection();
    }
}

void ConnectionManagerMainWindow::readDomains(QXmlStreamReader *xml)
{
    if (xml->atEnd())
        return;

    while (!(xml->isEndElement() && xml->name() == "Domains") && !xml->atEnd())
    {
        xml->readNext();
        if (xml->isStartElement() && xml->name() == "Domain")
        {
            domains.append(xml->readElementText());
        }
    }
}

void ConnectionManagerMainWindow::readAutoLaunchedApps(QXmlStreamReader *xml)
{
    if (xml->atEnd())
        return;

    while (!(xml->isEndElement() && xml->name() == "AutoLaunchedApps") && !xml->atEnd())
    {
        xml->readNext();
        if (xml->isStartElement() && xml->name() == "App")
        {
            autoLaunchedApps.append(xml->readElementText());
        }
    }
    if (!autoLaunchedApps.empty())
        autoLaunchConnectionName = autoLaunchedApps.at(0);
}

void ConnectionManagerMainWindow::checkAutoLaunch(QListWidgetItem *cur, QListWidgetItem *prev)
{
    if (cur == NULL)
        return;

    currentConnectionName = cur->text();
    if (currentConnectionName == autoLaunchConnectionName)
    {
        autoLaunchCheckBox->setChecked(true);
    }
    else
    {
        autoLaunchCheckBox->setChecked(false);
    }
}

void ConnectionManagerMainWindow::setAutoLaunchConnection(bool set)
{
    // check if there is the autolaunch config file in the current directory
    if (set)
    {
        autoLaunchConnectionName = connList->currentItem()->text();
    }
    else
    {
        autoLaunchConnectionName = "";
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::createMenus()
{
    fileMenu = menuBar()->addMenu(tr("&File"));
    fileMenu->addAction(changeDirAction);
    fileMenu->addSeparator();
    fileMenu->addAction(exitAction);

    connMenu = menuBar()->addMenu(tr("&Connection"));
    connMenu->addAction(newAction);
    connMenu->addAction(editAction);
    connMenu->addSeparator();
    connMenu->addAction(startAction);
    connMenu->addSeparator();
    connMenu->addAction(deleteAction);

    menuBar()->addSeparator();

    helpMenu = menuBar()->addMenu(tr("&Help"));
    helpMenu->addAction(helpAction);

}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::createToolBars()
{
    toolBar = addToolBar(tr("Actions"));
    toolBar->setFloatable(false);
    toolBar->setMovable(false);
    toolBar->setToolButtonStyle(Qt::ToolButtonTextUnderIcon);
    toolBar->addAction(newAction);
    toolBar->addAction(editAction);
    toolBar->addAction(deleteAction);
    toolBar->addSeparator();
    toolBar->addAction(startAction);
    toolBar->addSeparator();
    toolBar->addAction(helpAction);
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::createActions()
{
    newAction = new QAction(QIcon(":/icons/new.png"), tr("&New"), this);
    newAction->setShortcut(tr("Ctrl+N"));
    newAction->setStatusTip(tr("Create a new connection"));
    newAction->setToolTip(tr("Create connection"));
    connect(newAction, SIGNAL(triggered()), this, SLOT(newConnection()));

    editAction = new QAction(QIcon(":/icons/edit.png"), tr("&Edit"), this);
    editAction->setShortcut(tr("Ctrl+E"));
    editAction->setStatusTip(tr("Edit connection"));
    editAction->setToolTip(tr("Edit connection"));
    connect(editAction, SIGNAL(triggered()), this, SLOT(editConnection()));

    changeDirAction = new QAction(QIcon(":/icons/chdir.png"), tr("Current folder"), this);
    changeDirAction->setStatusTip(tr("Change current directory"));
    changeDirAction->setToolTip(tr("Change current directory"));
    connect(changeDirAction, SIGNAL(triggered()), this, SLOT(changeDir()));

    startAction = new QAction(QIcon(":/icons/start.png"), tr("S&tart"), this);
    startAction->setShortcut(tr("Ctrl+T"));
    startAction->setStatusTip(tr("Launch connection"));
    startAction->setToolTip(tr("Launch connection"));
    connect(startAction, SIGNAL(triggered()), this, SLOT(startConnection()));

    deleteAction = new QAction(QIcon(":/icons/delete.png"), tr("&Delete"), this);
    deleteAction->setShortcut(tr("Ctrl+D"));
    deleteAction->setStatusTip(tr("Delete connection"));
    deleteAction->setToolTip(tr("Delete connection"));
    connect(deleteAction, SIGNAL(triggered()), this, SLOT(deleteConnection()));

    exitAction = new QAction(QIcon(":/icons/exit.png"), tr("E&xit"), this);
    exitAction->setShortcut(tr("Alt+X"));
    exitAction->setStatusTip(tr("Exit"));
    connect(exitAction, SIGNAL(triggered()), this, SLOT(aboutToQuit()));

    helpAction = new QAction(QIcon(":/icons/help.png"), tr("&Help"), this);
    helpAction->setShortcut(tr("Ctrl+H"));
    helpAction->setStatusTip(tr("Show help"));
    connect(helpAction, SIGNAL(triggered()), this, SLOT(showHelp()));
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::contextMenuEvent(QContextMenuEvent *event)
{
    QMenu menu(this);
    menu.addAction(startAction);
    menu.addAction(editAction);
    menu.addSeparator();
    menu.addAction(deleteAction);
    menu.exec(event->globalPos());
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::closeEvent(QCloseEvent *event)
{
    if (QMessageBox::question(this, this->windowTitle(),
                          tr("Are you sure you want to exit?"),
                          QMessageBox::Yes | QMessageBox::No, QMessageBox::Yes) == QMessageBox::Yes)
    {
        event->accept();
    }
    else
    {
        event->ignore();
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::changeDir()
{
    QString directory = QFileDialog::getExistingDirectory(this,
                        tr("Find Files"), QDir::currentPath());
    if (directory == "") return;
    label->setText("Connection List (" + directory + ")");
    QDir::setCurrent(directory);
    QStringList files;
    QString fileName = "*.rdp";
    QDir dir = QDir(directory);
    files = dir.entryList(QStringList(fileName), QDir::Files | QDir::NoSymLinks);

    connList->clear();

    // If there are any entries in the list, enable the controls.
    if (files.size() > 0 ) enableConnectionActions();

    int i = 0;
    for (; i < files.size(); i++)
    {
        files[i].truncate(files[i].length() - 4);
        connList->addItem(files[i]);
    }
    workingDirName = directory;
}
//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::newConnection()
{
    ConnectionWizard *wizard = new ConnectionWizard(this);
    if (wizard->exec() == QDialog::Accepted)
    {
        QStringList files;
        QString fileName = "*.rdp";
        QDir dir = QDir(QDir::currentPath());
        files = dir.entryList(QStringList(fileName), QDir::Files | QDir::NoSymLinks);

        // Since there was a successful addition to the list, enable the controls
        enableConnectionActions();

        connList->clear();
        int i = 0;
        for (; i < files.size(); i++)
        {
            files[i].truncate(files[i].length() - 4);
            connList->addItem(files[i]);
        }
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::editConnection()
{
    if (connList->currentItem() != NULL)
    {
        QString pitFileName = connList->currentItem()->text() + ".rdp";
        ConnectionEditor *editor = new ConnectionEditor(pitFileName, this);
        if (editor->exec() == QDialog::Accepted)
        {
            QStringList files;
            QString fileName = "*.rdp";
            QDir dir = QDir(QDir::currentPath());
            files = dir.entryList(QStringList(fileName), QDir::Files | QDir::NoSymLinks);

            // Since there was a successful addition to the list, enable the controls
            enableConnectionActions();

            connList->clear();
            int i = 0;
            for (; i < files.size(); i++)
            {
                files[i].truncate(files[i].length() - 4);
                connList->addItem(files[i]);
            }
        }
    }
    else
    {
        QMessageBox::critical(this, "Edit connection", "There are no connections selected");
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::startConnection()
{

    // First check to see that the list is not empty
    if ( (connList->count()) == 0)
    {
        QMessageBox::critical(this, "Start Connection", "The connection list is empty.");
        return;
    }

    // Check that an entry is actually selected
    if ( connList->currentItem() == NULL )
    {
        QMessageBox::critical(this, "Start connection", "No connection has been selected.");
        return;
    }

    QString fname = currentConnectionName = connList->currentItem()->text();
    QFileInfo fi(fname + ".rdp");
    QFile file(fi.absoluteFilePath());
    if (!file.open(QIODevice::ReadOnly | QIODevice::Text))
    {
        QMessageBox::information(this, "Start connection - " + fname, "Error: can't open the connection file");
        return;
    }
    QTextStream stream(&file);
    QString line, name, val;
    QString brokerPort, brokerProtocol;
    QString listType;
    QString wait, retry;
    int listMask = 0;
    while (!stream.atEnd())
    {
        name = line = stream.readLine();
        if (line == "") break;
        int pos = line.indexOf(":");
        name.truncate(pos);
        val = line.remove(0, pos + 3);
        if (name == "username")
            username = val;
        else if (name == "password")
            password = val;
        else if (name == "domain")
            domain = val;
        else if (name == "alternate shell")
            appName = val;
        else if (name == "llmservers")
            broker = val;
        else if (name == "llmtcpport")
            brokerPort = val;
        else if (name == "llmprotocol")
            brokerProtocol = (val == "HTTPS"? "https://" : "http://");
        else if (name == "grantApps" && val != "0")
            listMask |= 1;
        else if (name == "grantDesktops" && val != "0")
            listMask |= 2;
        else if (name == "wait")
            wait = val;
        else if (name == "retry")
            retry = val;
    }
    if (brokerPort != "")
    {
        broker.append(QString(":") + brokerPort) ;
    }
    broker.prepend(brokerProtocol);
    if (listMask == 1) listType = "-La";
    else if (listMask == 2) listType = "-Ld";
    else listType = "-L";
    if (username == "" || password == "" || domain == "")
    {
        CredentialsForm form(this);
        if (form.exec() == QDialog::Rejected)
        {
            return;
        }
    }
    if (password.length() == 512)
    {
        unsigned char buffer[258];
        char passw[64];
        ::HexToBin(buffer, sizeof(buffer), password.toLatin1().data(), 512);
        ::Decrypt(buffer, sizeof(buffer));
        ::Decode256(buffer, passw);
        password = passw;
    }
    if (appName == "")
    {
        ApplicationForm form(this, broker, username, password, domain, listType);
        if (form.exec() == QDialog::Rejected)
        {
            return;
        }
    }
    if (appName.length() == 2048)
    {
        unsigned char buffer[1026];
        char shell[256];
        ::HexToBin(buffer, sizeof(buffer), appName.toLatin1().data(), 2048);
        ::Decrypt(buffer, sizeof(buffer));
        ::Decode1024(buffer, shell);
        appName = shell;
    }
    file.close();

    startAction->setEnabled(false);

    // we have now credentials, broker address and application name,
    // everything else are contained in the pit-file.
    currentStatusMessage = " - connection in progress...";
    pntsc = new QProcess;
    connect (pntsc, SIGNAL(readyReadStandardError()),
             this, SLOT(readChildErrors()));
    connect (pntsc, SIGNAL(readyReadStandardOutput()),
             this, SLOT(readChildOutput()));

    pntsc->start("pntsc", QStringList() << broker << "-u" << ( username )
                 << "-p" << ( password ) << "-d" << ( domain )
                 << "-a" << appName << "-f" << fi.absoluteFilePath()
                 << "-w" << wait << "-n" << retry);
    // syslog( LOG_NOTICE, "ConnManager: u:%s p:%s d:%s.", username.toLatin1().data(), password.toLatin1().data(), domain.toLatin1().data() );
    if (!pntsc->waitForStarted(1000))
    {
        QMessageBox::critical(this,
            "Starting connection - " + currentConnectionName.toUpper(),
            "Unable to start PNTSC");
        pntsc->kill();
    }
    else
    {
        while (pntsc->state() == QProcess::Running)
        {
            statusBar()->showMessage(currentStatusMessage);
            QApplication::processEvents();
        }
    }
    delete pntsc;
    pntsc = NULL;
    startAction->setEnabled(true);
    statusBar()->showMessage("Ready");
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::readChildOutput()
{
    QByteArray data = pntsc->readAllStandardOutput();
    QStringList list = QString(data).split("\n", QString::SkipEmptyParts);

    if (list.first().startsWith("POWERING_UP"))
    {
        const int timeoutOffset = 1;
        const int attemptOffset = 2;
        const int msgOffset = 3;
        const QString defaultMsg(tr("The computer operating system is loading..."));
        const QString plsWaitMsg(tr(" Please wait."));

        QString message;
        int timeout = list.at(timeoutOffset).toInt();
        int curAttemptNum = list.at(attemptOffset).toInt();
        message = list.at(msgOffset);
        if (message.isEmpty())
        {
            message = defaultMsg;
        }
        message += plsWaitMsg;

        RestartProgressDialog *restartProgressDialog = RestartProgressDialog::getInstance(&pntsc, message, timeout, curAttemptNum);
        restartProgressDialog->show();
    }
    else if (!list.last().startsWith("- ") && list.last() != "")
    {
        QMessageBox::information(this,
            "Starting connection - " + currentConnectionName.toUpper(),
            list.last());
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::readChildErrors()
{
    QByteArray data = pntsc->readAllStandardError();

    // parse received error string
    QStringList list = QString(data).split("],");
    QString code     = list.at(0);
    QString module   = list.at(1);
    QString inf      = list.at(2);

    code.replace("[ERR","");
    module.replace("[","");

    bool ok;
    int errcode = code.toInt(&ok, 16);
    QString errcodestr = getPntscErrorString(errcode);

    QMessageBox::critical(this, "Getting an applications list",
        "[Error " + code + "]: " + errcodestr +
        "\n[Module]: " + module +
        "\n[Description]: " + inf);
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::deleteConnection()
{
    if (QMessageBox::question(this, this->windowTitle(),
                              tr("You are about to delete the settings for connection [") + (connList->currentItem()->text()) + "].\n"
                              + tr("This cannot be undone.\n")
                              + tr("Are you sure you want to continue?"),
                              QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
    {
        return;
    }

    if (connList->currentItem() != NULL)
    {
        QFile file(connList->currentItem()->text() + ".rdp");
        file.remove();
        connList->takeItem(connList->currentRow());

        // If the list is now empty disable the connection controls
        if ( connList->count() == 0 ) disableConnectionActions();
    }
    else
    {
        QMessageBox::critical(this, "Delete connection", "There are no connections selected");
    }
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::aboutToQuit()
{
    // write the XML config file
    QString path = QDir::homePath() + QDir::separator() + XML_CONFIG_PATH;
    QDir dir (path);
    if (!dir.exists())
	QDir().mkpath(path);

    // TODO: remove the next three lines some times in future
    QFile oldFile(QDir::homePath() + QDir::separator() + XML_CONFIG_FILE_NAME);
    if (oldFile.exists())
	oldFile.remove();

    QFile confFile(path + QDir::separator() + XML_CONFIG_FILE_NAME);
    if (!confFile.open(QIODevice::ReadWrite | QIODevice::Text))
    {
        QMessageBox::critical(this, APPLICATION_NAME, "Can't open the config file.");
    }
    confFile.resize(0);
    QXmlStreamWriter xml(&confFile);
    xml.setAutoFormatting(true);
    xml.writeStartDocument();

    xml.writeStartElement("config");
    xml.writeStartElement("WorkingDirectory");
    xml.writeCharacters(workingDirName);
    xml.writeEndElement();

    // for the time being save only the first autolaunched application
    xml.writeStartElement("AutoLaunchedApps");
    if (autoLaunchConnectionName != "")
    {
        xml.writeStartElement("App");
        xml.writeCharacters(autoLaunchConnectionName);
        xml.writeEndElement();
    }
    else
    {
        xml.writeEmptyElement("App");
    }
    xml.writeEndElement();

    // for the time being save only the first autolaunched application
    if (domains.count() > 0)
    {
        xml.writeStartElement("Domains");
        for (int i = 0; i < domains.count(); ++i)
        {
            xml.writeStartElement("Domain");
            xml.writeCharacters(domains[i]);
            xml.writeEndElement();
        }
        xml.writeEndElement();
    }
    else
    {
        xml.writeEmptyElement("Domains");
    }

    xml.writeEndDocument();

    confFile.close();
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::showHelp()
{
    QMessageBox::information(this, "Help - " APPLICATION_NAME, tr(APPLICATION_NAME "\n"
                             "Copyright (c) 2010 Quest Software.\n"
                             "http://www.quest.com"));
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::setCredentials(const QString& u,
        const QString& p,
        const QString& d)
{
    username = u; password = p;
    domain = d;
}

//----------------------------------------------------------------------------
void ConnectionManagerMainWindow::setApplicationName(const QString& app)
{
    appName = app;
}

QStringList ConnectionManagerMainWindow::domainList()
{
    return domains;
}

void ConnectionManagerMainWindow::addDomain(const QString& domain)
{
    if (domain != "" && !domains.contains(domain))
    {
        domains.append(domain);
    }
}

void ConnectionManagerMainWindow::enableConnectionActions( void )
{
    startAction->setEnabled( true );
    deleteAction->setEnabled( true );
    editAction->setEnabled( true );
}

void ConnectionManagerMainWindow::disableConnectionActions( void )
{
    startAction->setEnabled( false );
    deleteAction->setEnabled( false );
    editAction->setEnabled( false );
}


