
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


//
// Default includes
//
#include "../common/common.h"

//
// Local #includes
//
#include <exception>
#include <stdexcept>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <errno.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <popt.h>
#include "../common/simplexml.h"
#include "../common/crypto.h"
#include "../common/rc4.h"
#include "../common/errors.h"

#define SERVER_PORT 42332
#define PNTSC_PORT  42331
#define GUI_PORT    42330

// Ticket location
#define Q_SECURE_TICKET_FILE "/tmp/appportal/qrdesktop_tix.bin"
#define MAX_RESPONSE_SIZE    1024*1024

//TODO: remove it into utils namespace

static const char * usagestr =
    "\npntsc: Quest Software Virtual AccessClient Request Handler\n\
    Version 7.0.0.012 (GA) Copyright (C) 2004-2010 Quest Software.\n\
    http://www.vWorkspace.com\n\n\
    Usage: pntsc [OPTIONS]\n\
    For example:\n\
    pntsc http://broker.mycompany.com -u myusername -p mypassword -d mydomain -l\n\
    pntsc http://broker.mycompany.com:8080 -T /home/filename.tgt -s -a \"Virtual Desktop\" -o \"-a 16 -g 1024x768\"\n\
    pntsc -o\n\
    pntsc -f /home/user/Calculator\n\n\
    Options:";


class noncopyable
{
protected:
    noncopyable() {}
    ~noncopyable() {}
private:
    noncopyable( const noncopyable& );
    const noncopyable& operator=( const noncopyable& );
};
//-----------------------------------------------------------------------------
class CSocketGuard : noncopyable
{
public:
    void reset(int socket)
    {
        if (m_socket != -1)
        {
            close(m_socket);
        }
        m_socket = socket;
    }
    CSocketGuard(int socket = -1)
            : m_socket(-1)
    {
        reset(socket);
    }
    ~CSocketGuard()
    {
        reset(-1);
    }
    int get()
    {
        return m_socket;
    }
    int release()
    {
        int socket = m_socket;
        m_socket = -1;
        return socket;
    }
private:
    int m_socket;
};

//-----------------------------------------------------------------------------
class CPoptContextGuard : noncopyable
{
public:
    void reset(poptContext pCtx)
    {
        if (m_pCtx)
        {
            ::poptFreeContext(m_pCtx);
        }
        m_pCtx = pCtx;
    }

    CPoptContextGuard(poptContext pCtx = 0) : m_pCtx(0)
    {
        reset(pCtx);
    }

    ~CPoptContextGuard()
    {
        reset(0);
    }

    poptContext get()
    {
        return m_pCtx;
    }

    poptContext release()
    {
        poptContext tmp = m_pCtx;
        m_pCtx = 0;
        return tmp;
    }

private:
    poptContext m_pCtx;
};


//-----------------------------------------------------------------------------
int form_command_line(char *message, char **args)
{
    char *p, *q, *quot1, *quot2;
    int i;

    try
    {
        p = &message[5];
        // printing command line of application
        for (i = 0; i < 128; ++i)
        {
            if (!(q = strchr (p, '\n')) || (*p == 0))
            {
                break;
            }
            *q = '\0';
            // remove pair quotes from the parameter
            while (( quot1 = (strchr(p, '"'))))
            {
                quot2 = strrchr(p, '"');
                if (quot1 == quot2)
                {
                    sprintf (message, "Extra quotes in parameters string.");
                    throw std::exception();
                }
                else
                {
                    memmove(quot1, quot1+1, quot2-quot1-1);
                    memmove(quot2-1, quot2+1, strlen(quot2)-1);
                    p[strlen(p)-2] = '\0';
                }
            }

            args[i] = (char *) malloc (strlen(p)+1);
            strcpy (args[i], p);
            p = q+1;
        }
    }
    catch (const std::exception&)
    {
        while (i > 0)
        {
            free (args[--i]);
        }
        return 0;
    }

    if (i == 128)
    {
        sprintf (message, "Too many parameters to launch application.\n");
        while (i > 0)
        {
            free (args[--i]);
        }
    }
    else if (i == 0)
    {
        sprintf (message, "There is no launch information in server response");
    }

    args[i] = NULL;
    return i;
}

//-----------------------------------------------------------------------------
void sigusr1_handler(int signo)
{
    fprintf (stdout, "- Connection started\n");
    fflush (stdout);
    exit(0);
}

//-----------------------------------------------------------------------------
void ReadPITFile(const char *pitfilename, SimpleString& ssPitFile)
{
    int ret = 0;
    char errstr[1024];
    if (pitfilename == NULL)
    {
        throw std::runtime_error("RDP file name must be set.");
    }
    int pitfile = open (pitfilename, O_RDONLY);
    if (pitfile == -1)
    {
        sprintf (errstr, "Cannot open RDP file: %s", strerror (errno));
        throw std::runtime_error(errstr);
    }
    int pitfilesize = lseek (pitfile, 0L, SEEK_END);
    char *pitbuf = (char *) malloc (pitfilesize);
    lseek (pitfile, 0L, SEEK_SET);
    if ((ret = read (pitfile, pitbuf, pitfilesize)) < 0)
    {
        sprintf (errstr, "Error on reading RDP file: %s", strerror (errno));
        close (pitfile);
        free (pitbuf);
        throw std::runtime_error(errstr);
    }
    if (ret == 0)
    {
        ret = pitfilesize;
    }
    ssPitFile.LoadData (pitbuf, ret);
    close (pitfile);
    free (pitbuf);
}

//-----------------------------------------------------------------------------
// Return >0 if no problems, 0 if error.
int GetAndReadPitFile(const char *encryptedUrlPitFileName, SimpleString& ssPitFile)
{
    unsigned char encryptedUrl[4096];
    unsigned char url[4096];
    char* buffer;
    char command[1024];
    SimpleString strResponse;
    int ret;
    // int i;

    buffer = (char*)malloc(MAX_RESPONSE_SIZE);
    ret    = 0;

    // Decrypt the file
    if (qdc_crypto_decrypt_file( encryptedUrlPitFileName ) == 0)
        return 0;   // pitfilename if empty or the pit file is broken
    ReadPITFile( encryptedUrlPitFileName, ssPitFile);

    printf( "Reading initial pit file '%s'.\n", encryptedUrlPitFileName );

    // Look for the URL tag.  If it is here then get it, decrypt it
    // and then reload *that* file.
    if (ssPitFile.FindItem("url:s:",( char *) encryptedUrl) != 0 )
    {
        // Decrypt the URL string.
        printf( "GetAndReadPitFile(): The encrypted URL string is '%s'.\n\n", encryptedUrl );
        Decrypt1024( encryptedUrl, url, 4096);
        printf( "GetAndReadPitFile(): I decrypted the URL as '%s'.\n\n", url );

        // Generate the command string
        sprintf( command, "/usr/bin/wget --no-check-certificate --user-agent=\"Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)\" "\
                    " \"%s\" -O \"%s\"", url, encryptedUrlPitFileName);

        // Do it
        system( command );

        // Decrypt it locally
        qdc_crypto_decrypt_file( encryptedUrlPitFileName );
        ReadPITFile( encryptedUrlPitFileName, ssPitFile);

        ret = 1;
    }

    return ret;
}

int locateAndObscureArg( int argc, char ** argv,  char * option )
{
    int option_index;
    int i;
    char * ptr;

    // Find the index of the targetted option
    for ( i = 0; i < argc ; i ++ )
    {
        if (strcmp( argv[i], (char * )option ) == 0) 
        {
            option_index = i + 1;
            break;
        }
    }

    // Erase it
    // Sanity check first - do not try to do this if there aren't
    //                      enough parameters
    if ( option_index < argc )
    {
        ptr = argv[ option_index ];

        // Replace all the chars of the following option with 'x'
        while ( *ptr ) *(ptr ++) = 'x';
    }
}



//-----------------------------------------------------------------------------
int main(int argc, const char **argv)
{
    char message[65336];
    const char *broker;
    char *user = NULL;
    char *password = NULL;
    char *domain = NULL;
    char *app = NULL;
    char *sslgateway = NULL;
    char *pitfilename = NULL;
    char *tgtfilename = NULL;
    char *proxyaddr = NULL;
    char *options = NULL;
    char *user1 = NULL;
    char *password1 = NULL;
    char *domain1 = NULL;
    char *app1 = NULL;
    char *new_password = NULL;
    char *ticket = NULL;
    int list = 0, List = 0, lista = 0, Lista = 0, listd = 0, Listd = 0;
    // int list_type = 1;
    int nat = 0;
    int wait = 5;
    int retry = 20;
    // int ret = 0;
    int firstDesktop = 0;
    int ipOnly = 0;
    SimpleString ssPitFile;

    int o_opt = 0;
    int i, res = 0;

    char * encrypted_user;
    char * encrypted_password;
    char * encrypted_domain;
    char * opt_ptr;
    char s1[80];
    char errstr[1024];
//    char s2[80],s3[80],s4[80],s5[80],s6[80];
    char buf1[256];
    char buf2[256];
    char buf3[256];

    openlog( "pntsc", LOG_PID, LOG_USER);
    LOG_BANNER( "pntsc" );

    sprintf (buf1, "mkdir /tmp/appportal");
    system(buf1);

    poptContext optCon;   /* context for parsing command-line options */
    struct poptOption optionsTable[] =
    {
        {
            NULL, 'u', POPT_ARG_STRING, &user, 0, "user name (If not specified, the user will be prompted for user name)"
        },
        { NULL, 'p', POPT_ARG_STRING, &password, 0, "password (If not specified, the user will be prompted for password)" },
        { NULL, 'd', POPT_ARG_STRING, &domain, 0, "domain (If not specified, the user will be prompted for domain name)" },
        { NULL, 'l', POPT_ARG_NONE, &list, 0, "display cached list of published applications and desktops" },
        { NULL, 'L', POPT_ARG_NONE, &List, 0, "query broker for list of published applications and desktops" },
        { "la",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &lista, 0, "display cached list of published applications" },
        { "La",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &Lista, 0, "query broker for list of published applications" },
        { "ld",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &listd, 0, "display cached list of published desktops" },
        { "Ld",  0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &Listd, 0, "query broker for list of published desktops" },
        { "FD",   0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &firstDesktop, 0, "Find the first Desktop" },
        { "ip",   0,  POPT_ARG_NONE|POPT_ARGFLAG_ONEDASH, &ipOnly, 0, "Only return the IP address of the target host" },
        { NULL, 'a', POPT_ARG_STRING, &app, 0, "published application or desktop to launch" },
        { NULL, 'w', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &wait, 0, "retry interval" },
        { NULL, 'n', POPT_ARG_INT|POPT_ARGFLAG_SHOW_DEFAULT, &retry, 0, "number of retries" },
        { NULL, 't', POPT_ARG_NONE, &nat, 0, "enable NAT support for firewall traversal (for use with Terminal Services servers only)" },
        { NULL, 'o', POPT_ARG_STRING|POPT_ARGFLAG_OPTIONAL, &options, 1, "qrdesktop options" },
        { NULL, 'f', POPT_ARG_STRING, &pitfilename, 0, "pit-file name" },
        { NULL, 'e', POPT_ARG_STRING, &sslgateway, 0, "SSL gateway name" },
        { NULL, 'T', POPT_ARG_STRING, &tgtfilename, 0, "TGT-file name" },
        { NULL, 'c', POPT_ARG_STRING, &new_password, 0, "change current password to this" },
        { NULL, 'P', POPT_ARG_STRING, &proxyaddr, 0, "Proxy address (in the format 'address:port')" },
        POPT_AUTOHELP
        POPT_TABLEEND
    };

    optCon = poptGetContext(NULL, argc, argv, optionsTable, 0);
    poptSetOtherOptionHelp(optCon, usagestr);
    try
    {
        if (argc < 2) // at least 2 elements must be present in command line:
        {
            // pntsc -o
            poptPrintHelp(optCon, stderr, 0);
            sprintf (errstr, "[ERR%05X],[pntsc],Wrong command line",
                INPUT_PARAMETERS_ERROR);
            throw std::runtime_error(errstr);
        }
        CPoptContextGuard poptGuard(optCon);
        // poptGetNextOpt returns 1 after -o option had been parsed,
        // -1 when last parameter had been parsed,
        // and some another negative value if an error occured.
        if ((res = poptGetNextOpt(optCon)) == 1)
        {
            o_opt = 1;
            res = poptGetNextOpt(optCon);
        }
        if (res != -1)
        {
            poptPrintHelp(optCon, stderr, 0);
            throw std::exception();
        }

        broker = poptGetArg(optCon);
        if (broker == NULL)
        {
            if (pitfilename == NULL)
            {
                if (argc != 2 || o_opt == 0)
                {
                    poptPrintHelp(optCon, stderr, 0);
                    throw std::exception();
                }
                else
                {
                    execlp("qrdesktop", "qrdesktop", 0, (char *) 0 );
                    exit (0);
                }
            }
        }
        else if ((strncmp(broker, "http://", 7) && strncmp(broker, "https://", 8)) ||
                 poptPeekArg(optCon) != NULL)    // if more than 1 leftover arguments are present
        {
            poptPrintHelp(optCon, stderr, 0);
            throw std::exception();
        }

        list = list || List || lista || Lista || listd || Listd || firstDesktop;

        if ( user != NULL )
         {
                 locateAndObscureArg( argc, (char **)argv, "-u" );
         }
        if ( domain != NULL )
         {
                 locateAndObscureArg( argc, (char **)argv, "-d" );
         }
        if ( password != NULL )
         {
                 locateAndObscureArg( argc, (char **)argv, "-p" );
         }

        // the following is based on the assumption that we always get either
        // credentials or pitfilename with either credentials or ticket from GUI

        // if we do not get neither of them, then we assume working
        // from the command-line, and read credentials from stdin

        // now we have either correct broker address or pitfile name
        if ( user == NULL || password == NULL || domain == NULL)
        {
            int pitFileLength;
            ticket = (char *) malloc(4096);

            syslog( LOG_NOTICE, "Credential set incomplete." );

            // preparing the pit-file data
            pitFileLength = GetAndReadPitFile(pitfilename, ssPitFile);  // throws an std::exception on any error

            if ( pitFileLength != 0)
            {
                // now let's try to find credentials in the RDP file
                // if something is absent, request it from stdin
                // but only if no ticket is present
                printf(  "Looking for ticket info in %s\n", pitfilename);

                if (ssPitFile.FindItem("ticket:s:", ticket) != 0 && *ticket != 0 )
                {
                    FILE * ofile;

                    // DCT: save the ticket data into a tgt file
                    printf(  "Saving ticket info in %s\n", Q_SECURE_TICKET_FILE);
                    ofile = fopen( Q_SECURE_TICKET_FILE, "wb" );
                    fwrite(( char * )ticket, 1, strnlen( (char *)ticket, 4096), ofile );
                    fclose( ofile );
                }
            }

            // Only if the ticket is not found, then get the rest of the credentials
            // Implicitly, if the ticket *IS* there, no other credentials will be present
            if ( strnlen(ticket, 4096) == 0 )
            {
            if (user == NULL)
            {
                user1 = (char *) malloc(128);
                if (ssPitFile.FindItem("username:s:", user1) == 0 || *user1 == 0)
                {
                    printf("username: ");
                    if (scanf("%128s", user1) < 1)
                    {
                        throw std::exception();
                    }
                }
            }
            if (domain == NULL)
            {
                domain1 = (char *) malloc(128);
                if (ssPitFile.FindItem("domain:s:", domain1) == 0 || *domain1 == 0)
                {
            	    char *p = strchr(user1, '@');
            	    if (p)
            	    {
            		// get domain from username
            		strcpy(domain1, p);
            		*p = '\0';
            	    }
            	    else
            	    {
                    printf("domain: ");
                    if (scanf("%128s", domain1) < 1)
                    {
                        throw std::exception();
                    }
                }
            }
            }
            if (password == NULL)
            {
            	char szPassword[514];
            	if (ssPitFile.FindItem("password:s:", szPassword) != 0 && *szPassword != 0)
            	{
                    unsigned char rgbBuffer[257];
                    password1 = (char *) malloc(128);
                    // Decrypt the password.
                    HexToBin(rgbBuffer, sizeof(rgbBuffer), szPassword, strlen(szPassword));
                    Decrypt(rgbBuffer, sizeof (rgbBuffer));
                    Decode256(rgbBuffer, password1);
            	}
            	else
            	{
                    if ((password1 = getpass("password: ")) == NULL)
                    {
                        fprintf(stderr, "Error: %s", strerror(errno));
                        throw std::exception();
                    }
                }
            }
            }
        }
        if (!list && !new_password && app == NULL)
        {
            // check if the 'alternate shell' field is set in the RDP file
            if (ssPitFile.Length() == 0)
            {
                // preparing the pit-file data
                ReadPITFile(pitfilename, ssPitFile);  // throws an std::exception on any error
            }
            char szApp[2050];
            if (ssPitFile.FindItem("alternate shell:s:", szApp) == 0 || *szApp == 0)
            {
                fprintf (stderr, "[ERR%05X],[pntsc],Alternate shell field is not set in the RDP file.\n",
                    INPUT_PARAMETERS_ERROR);
                throw std::exception();
            }

        }

        // now we have credentials
        if ((!list && !new_password && app == NULL) && pitfilename == NULL && tgtfilename == NULL)
        {
            poptPrintHelp(optCon, stderr, 0);
            throw std::exception();
        }
    }
    catch (const std::runtime_error & e)
    {
        fprintf (stderr, "[ERR%05X],[pntsc],%s", INTERNAL_ERROR, e.what());
        if (user1 != NULL) free (user1);
        if (password1 != NULL) free (password1);
        if (domain1 != NULL) free (domain1);
        if (app1 != NULL) free (app1);
        if (ticket != NULL ) free (ticket);
        exit(1);
    }
    catch (const std::exception&)
    {
    }
    catch (...)
    {
        fprintf (stderr, "[ERR%05X],[pntsc],An unknown exception was caught - check your connection parameters", INTERNAL_ERROR);
    }
//-----------------------------------------------------------------------------

// sending:

#ifdef DEBUG
    syslog (LOG_NOTICE, "creds: %s %s %s\n", user1, password1, domain1);
    syslog (LOG_NOTICE, "creds: %s %s %s\n", user, password, domain);
#endif

    try
    {
        char err[] = "Starting connection management server failed";
        // connect to the server
        int sock = socket (AF_INET, SOCK_STREAM, 0);
        if (sock < 0)
        {
            throw std::runtime_error(err);
        }

        CSocketGuard sockGuard(sock);

        int val = 1;
        setsockopt (sock, SOL_SOCKET, SO_REUSEADDR, &val, sizeof (val));
        struct sockaddr_in addr;
        bzero (&addr, sizeof (addr));
        addr.sin_family = AF_INET;
        addr.sin_port = htons(SERVER_PORT);
        addr.sin_addr.s_addr = inet_addr ("127.0.0.1");
        if (connect (sock, (struct sockaddr *) &addr, sizeof (addr)) < 0)
        {
            // try to start the server
            fprintf (stdout, "- Starting control server...\n");
            fflush (stdout);
            if ((res = fork()) < 0)
            {
                throw std::runtime_error(err);
            }
            else if (res == 0)
            {
                sockGuard.reset(-1);
                if (execlp ("pncm", "pncm", (char*)0))
                {
                    throw std::runtime_error(err);
                }
                exit (0);
            }
            sleep (1);
            if (connect (sock, (struct sockaddr*) &addr, sizeof (addr)))
            {
                throw std::runtime_error(err);
            }

        }   // if (connect < 0)

        // form a message to the server:
        // transform command line into a string and pass it to the server
        strcpy(message, "#PNTSC#");
        for (i = 1; i < argc; ++i)
        {
            if (strchr(argv[i], ' '))
            {
                strcat(message, "\"");
                strcat(message, argv[i]);
                strcat(message, "\"");
            }
            else
            {
                strcat(message, argv[i]);
            }
            strcat(message, " ");
        }
        if (user1 != NULL)
        {
            strcat(message, "-u ");
            strcat(message, user1);
            strcat(message, " ");
        }
        if (password1 != NULL)
        {
            strcat(message, "-p ");
            strcat(message, password1);
            strcat(message, " ");
        }
        if (domain1 != NULL)
        {
            strcat(message, "-d ");
            strcat(message, domain1);
            strcat(message, " ");
        }
        if (user != NULL)
        {
            strcat(message, "-u ");
            strcat(message, user);
            strcat(message, " ");
        }
        if (password != NULL)
        {
            strcat(message, "-p ");
            strcat(message, password);
            strcat(message, " ");
        }
        if (domain != NULL)
        {
            strcat(message, "-d ");
            strcat(message, domain);
            strcat(message, " ");
        }

        fprintf (stdout, "- Sending message to the server...\n");
        // syslog ( LOG_NOTICE , "- Sending message to the server...'%s'\n", message);
        fflush (stdout);
        write (sock, message, strlen(message));
        sleep(1);
        fprintf (stdout, "- Waiting for the server response...\n");

        // If it is a list request, then add the extra line feed
        if ( list ) fprintf( stdout, "\n" );

        fflush (stdout);
        sleep(1);
        do
        {
            if ((res = read (sock, message, sizeof (message) - 1)) < 0)
            {
                throw std::runtime_error("Read error");
            }
            else if (res == 0)
            {
                throw std::runtime_error("Connection to server was unexpectedly broken");
            }
            else
            {
                message[res] = 0;
            }
            // if not PENDING has been returned, go on
            if (strncmp (message, "PENDING", 7))
            {
                break;
            }
            else
            {
                fprintf (stdout, "%s", message + 7);
                fflush (stdout);
            }
        }
        while (1);

// end_sending:
        ;

    }
    catch (std::runtime_error& exc)
    {
        fprintf(stderr, "[ERR%05X],[pntsc],%s: %s",
            SYSTEM_ERROR, exc.what(), strerror(errno));
        if (user1 != NULL) free (user1);
        if (password1 != NULL) free (password1);
        if (domain1 != NULL) free (domain1);
        if (ticket != NULL ) free (ticket);
        exit(1);
    }

    if (user1 != NULL) free (user1);
    if (password1 != NULL) free (password1);
    if (domain1 != NULL) free (domain1);

    if (!strncmp (message, "START", 5))
    {
        // forming command line from given launch information
        char *args[128];
        int i;

        if ( ipOnly )
        {
            // Parse the IP address out and dump it to stdout
            sscanf( message, "%*s\n%*s\n%*s\n%*s\n%s\n%*s\n", s1 );
            printf( "%s\n", s1);
            exit (0);
        }

        if ((i = form_command_line(message, args)) == 0)
        {
            fprintf (stderr, "[ERR%05X],[pntsc],Unable to start application: %s",
                INPUT_PARAMETERS_ERROR, message);
            goto end;
        }

        fprintf (stdout, "- Starting the connection...\n");
        fflush (stdout);

        if ((i = fork()) < 0)
        {
            perror ("Can't create process");
            exit (1);
        }
        else if (i)
        {
            struct sigaction sa;
            sigemptyset(&sa.sa_mask);
            sa.sa_handler = sigusr1_handler;
            sigaction(SIGUSR1, &sa, NULL);

            pid_t pid;
            int stat;
            while ((pid = waitpid(-1, &stat, 0)) > 0);
            fprintf (stderr, "[ERR%05X],[pntsc], Rdesktop error - Unable to connect to server.\n",
			SYSTEM_ERROR );

//            fprintf(stderr, "Rdesktop error - Unable to connect to server.\n");
            exit(1);
        }
        setsid();
        close (0);
        close (1);
        close (2);
        if ((i = open("/dev/null", O_RDWR)) != -1)
        {
            dup(i);
            dup(i);
        }

        if ( ! firstDesktop )
        {
            printf( "Attempting to launch process.\n");

            // launch application and exit
            if (execvp (args[0], args))
            {
                perror ("Connection failed");
                exit (1);
            }
        }
    }
    else if (!strncmp(message, "[ERR", 4))
    {
        fprintf (stderr, "%s\n", message);
        fflush (stderr);
        return -1;
    }
    else if (strncmp(message, "SHARED", 6))
    {
        fprintf (stdout, "%s\n", message);
        fflush (stdout);
    }
    // and send nothing when some session was shared


end:

    fflush( stderr );
    fflush( stdout );

    return 0;
}

