
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include "rdesktop.h"   // for error() and DEBUG_PNRDP
#include <jpeglib.h>
#include <jerror.h>
#include <assert.h>
#include <setjmp.h>

struct jpeg_decompress_struct g_cinfo;

typedef struct
{
    struct jpeg_source_mgr pub;	// public fields

    unsigned char * cdata;		// the pointer to the compressed data
    size_t cdatasize;        // the size of the data
    JOCTET * buffer;		// the input buffer
} pnrdp_source_mgr;

typedef struct
{
    struct jpeg_error_mgr pub;	// "public" fields

    jmp_buf setjmp_buffer;	// for return to caller
} pnrdp_error_mgr;

pnrdp_error_mgr g_jerr;

#define JPEG_MAX_SCANLINES 4

// a custom error handler
METHODDEF(void)
pnrdp_error_exit (j_common_ptr cinfo)
{
    DEBUG_PNRDP(("%s\n", __func__));
    assert(cinfo != NULL);
    
    pnrdp_error_mgr *err = (pnrdp_error_mgr *) cinfo->err;

    // Display the message.
    (*cinfo->err->output_message) (cinfo);

    // Return control to the setjmp point
    longjmp(err->setjmp_buffer, 1);
}

// a custom output message handler
METHODDEF(void)
pnrdp_output_message(j_common_ptr cinfo)
{
    DEBUG_PNRDP(("%s\n", __func__));

    char buffer[JMSG_LENGTH_MAX];

    (*cinfo->err->format_message) (cinfo, buffer);
    error(buffer);
}

METHODDEF(void)
init_source (j_decompress_ptr cinfo)
{
    DEBUG_PNRDP(("%s\n", __func__));

    // no work necessary here
}

METHODDEF(boolean)
fill_input_buffer (j_decompress_ptr cinfo)
{
    DEBUG_PNRDP(("%s\n", __func__));

    pnrdp_source_mgr * src = (pnrdp_source_mgr *) cinfo->src;
    if (src->cdata == NULL || src->cdatasize == 0)
    {
        ERREXIT(cinfo, JERR_INPUT_EMPTY);
    }

    memcpy(src->buffer, src->cdata, src->cdatasize);

    src->pub.next_input_byte = src->buffer;
    src->pub.bytes_in_buffer = src->cdatasize;
    DEBUG_PNRDP((" %p %p %d\n", src->buffer, src->pub.next_input_byte, src->pub.bytes_in_buffer));

    return TRUE;
}

METHODDEF(void)
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
    DEBUG_PNRDP(("%s\n", __func__));
    
    pnrdp_source_mgr * src = (pnrdp_source_mgr *) cinfo->src;
    DEBUG_PNRDP((" %p %p %d, to %d\n", src->buffer, src->pub.next_input_byte, src->pub.bytes_in_buffer, num_bytes));
    if (src->cdata == NULL || src->cdatasize == 0)
    {
        ERREXIT(cinfo, JERR_INPUT_EMPTY);
    }
    if (src->pub.bytes_in_buffer < num_bytes)
    {
        ERREXIT(cinfo, JERR_INPUT_EOF);
    }
    src->pub.next_input_byte += num_bytes;
    src->pub.bytes_in_buffer -= num_bytes;
    DEBUG_PNRDP((" %p %p %d\n", src->buffer, src->pub.next_input_byte, src->pub.bytes_in_buffer));
}


METHODDEF(void)
term_source (j_decompress_ptr cinfo)
{
    DEBUG_PNRDP(("%s\n", __func__));

    // no work necessary here
}

GLOBAL(int)
pnrdp_init_jpeg_library()
{
    DEBUG_PNRDP(("%s\n", __func__));
    pnrdp_source_mgr * src;

    g_cinfo.err = jpeg_std_error(&g_jerr.pub);
    g_jerr.pub.error_exit = pnrdp_error_exit;
    g_jerr.pub.output_message = pnrdp_output_message;

    jpeg_create_decompress(&g_cinfo);

    g_cinfo.src = (struct jpeg_source_mgr *)
                  (*g_cinfo.mem->alloc_small) ((j_common_ptr) &g_cinfo, JPOOL_PERMANENT,
                                               sizeof(pnrdp_source_mgr));
    src = (pnrdp_source_mgr *) g_cinfo.src;

    src->pub.init_source = init_source;
    src->pub.fill_input_buffer = fill_input_buffer;
    src->pub.skip_input_data = skip_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart; // use default method
    src->pub.term_source = term_source;
    src->cdata = NULL;
    src->cdatasize = 0;

    return 0;
}

GLOBAL(void)
pnrdp_free_jpeg_library()
{
    DEBUG_PNRDP(("%s\n", __func__));
    jpeg_destroy_decompress(&g_cinfo);
}

GLOBAL(int)
pnrdp_decompress_image (unsigned char *cdata, size_t cdatasize, void **buffer)
{
    DEBUG_PNRDP(("%s\n", __func__));
    if (g_cinfo.src == NULL)
    {
        error("JPEG decomressor is not initialized!\n");
        return -1;
    }
    if(cdata == NULL || cdatasize == 0)
    {
        ERREXIT(&g_cinfo, JERR_INPUT_EMPTY);
    }

    pnrdp_source_mgr * src = (pnrdp_source_mgr *) g_cinfo.src;
    src->cdata = cdata;
    src->cdatasize = cdatasize;
    src->buffer = (JOCTET *) malloc (src->cdatasize);
    src->pub.bytes_in_buffer = 0;
    src->pub.next_input_byte = NULL;
    if (src->buffer == NULL)
    {
	ERREXIT(&g_cinfo, JERR_OUT_OF_MEMORY);
    }

    // save the local stack environment so as to return here from an error handler
    if (setjmp(g_jerr.setjmp_buffer))
    {
        // If we get here, the JPEG code has signaled an error.
        // Just return now (we've already output the error message).
        if (src->buffer) free(src->buffer);
	return -1;
    }
    
    jpeg_read_header(&g_cinfo, TRUE);

    jpeg_start_decompress(&g_cinfo);

    int row_stride = g_cinfo.output_width * g_cinfo.output_components;

    *buffer = malloc (row_stride * g_cinfo.output_height);
    if (*buffer == NULL)
    {
	ERREXIT(&g_cinfo, JERR_OUT_OF_MEMORY);
    }

    // calc recommended output buffer height and allocating the memory
    int scnum = g_cinfo.rec_outbuf_height > JPEG_MAX_SCANLINES ?
                JPEG_MAX_SCANLINES :
                g_cinfo.rec_outbuf_height;
    int i, j;
    JSAMPLE *lines[JPEG_MAX_SCANLINES];
    for (i = 0; i < scnum; ++i)
    {
        lines[i] = (JSAMPLE *) malloc (row_stride);
        if (lines[i] == NULL)
        {
	    ERREXIT(&g_cinfo, JERR_OUT_OF_MEMORY);
        }
    }

    // reading scanlines
    i = 0;

    while (i < g_cinfo.output_height)
    {
        jpeg_read_scanlines(&g_cinfo, lines, scnum);
        for (j = 0; j < scnum; ++j)
        {
            memcpy(*buffer + row_stride * (i + j), lines[j], row_stride);
        }
        i += scnum;
    }

    jpeg_finish_decompress(&g_cinfo);

    // cleanup
    for (i = 0; i < scnum; ++i)
    {
        free(lines[i]);
    }
    free (src->buffer);

    return 0;
}
