
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include "rdesktop.h"

#include <speex/speex.h>
#include <speex/speex_stereo.h>

#include "qaudio_driver.h"

#include "qaudio_codec.h"

/*-----------------------speex encoder -------------------------------------------*/

const SpeexMode * GetMode ( int sampling )
{
    if ( sampling >= 25000 )
	return &speex_uwb_mode;
    if ( sampling >= 12500 )
	return &speex_wb_mode;
    return &speex_nb_mode;
}

uint32 codec_get_encoder_frame_samples ( speex_encoder * codec )
{
    uint32 frame_samples;
    speex_encoder_ctl( codec->encoder , SPEEX_GET_FRAME_SIZE, &frame_samples );

    return frame_samples * codec->format.nChannels;
}

uint32 codec_get_encoder_frame_bytes ( speex_encoder * codec )
{
    return sizeof(short) * codec_get_encoder_frame_samples( codec );
}

speex_encoder * codec_create ( LPWAVEFORMATEX format, int complexity, int quality )
{
    speex_encoder * codec = NULL;

    if ( format == NULL )
    {
		error ("CODEC: Invalid format\n");
		return NULL;
    }

    if (!( 0 < format->nChannels && format->nChannels <= 2 ))
    {
		error ("CODEC: Invalid format, channels count\n");
		return NULL;
    }

    codec = (speex_encoder*)xmalloc( sizeof ( speex_encoder ) );
    if ( codec == NULL )
    {
    	return NULL;
    }

    memcpy ( &codec->format , format, sizeof ( WAVEFORMATEX ) );

    codec->encoder = speex_encoder_init ( GetMode(format->nSamplesPerSec) );

    if ( codec->encoder == NULL )
    {
		error ( "CODEC: Can't create encoder\n");
		xfree(codec);
		return 0;
    }
    else
    {
	    speex_encoder_ctl(codec->encoder,SPEEX_SET_SAMPLING_RATE,&format->nSamplesPerSec);

        speex_encoder_ctl(codec->encoder, SPEEX_SET_COMPLEXITY, &complexity);

		speex_encoder_ctl(codec->encoder,SPEEX_SET_QUALITY, &quality);

		codec->data_buffer = (short*)xmalloc( codec_get_encoder_frame_bytes(codec) );
		codec->data_buffer_size = 0;

		speex_bits_init ( &codec->bits );
    }
    return codec;
}

void codec_destroy ( speex_encoder * codec )
{
    if ( codec == NULL )
    	return;

    if ( codec->encoder != NULL )
        speex_encoder_destroy( codec->encoder );

    if ( codec->data_buffer != 0 )
    	xfree ( codec->data_buffer );

    speex_bits_destroy ( &codec->bits );

    xfree( codec );
}


int codec_read_frame_16bps ( speex_encoder * codec , char * buffer , uint32 size , uint32 * offset , uint32 to_read_samples )
{
    uint32 to_read_bytes = to_read_samples * sizeof(short);
    uint32 avaliable_bytes = size - *offset;
    uint32 to_copy_bytes = (avaliable_bytes < to_read_bytes) ? avaliable_bytes : to_read_bytes;

    if ( to_copy_bytes == 0 ) return 0;

    memcpy ( codec->data_buffer + codec->data_buffer_size/sizeof(short) , buffer + *offset , to_copy_bytes );

    codec->data_buffer_size += to_copy_bytes ;
    *offset += to_copy_bytes;

    return ( to_copy_bytes == to_read_bytes ) ? 1 : 0;
}

int codec_read_frame_8bps  ( speex_encoder * codec , char * buffer , uint32 size , uint32 * offset , uint32 to_read )
{
    uint32 i;
    short * db = codec->data_buffer + codec->data_buffer_size / sizeof(short);

    uint32 to_copy = ( (size - *offset) < to_read )? (size - *offset) : to_read;

    buffer += *offset;

    for ( i = 0; i < to_copy; ++ i )
    {
		short val = buffer[i];
		val = ( val << 8 ) ^ 0x8000;
		*db++ = val;
    }

    codec->data_buffer_size += to_copy * sizeof(short);
    *offset += to_copy;

    return ( to_copy == to_read )? 1: 0;
}

int codec_read_frame ( speex_encoder * codec, char * buffer , uint32 size , uint32 * offset )
{
    uint32 to_read_samples = 0;
    uint32 frame_samples = codec_get_encoder_frame_samples ( codec );

    if ( codec->format.wBitsPerSample == 16 )
	{
		to_read_samples = frame_samples - codec->data_buffer_size / sizeof(short);
		return codec_read_frame_16bps ( codec , buffer , size , offset , to_read_samples );
	}
    else
	{
		to_read_samples = frame_samples - codec->data_buffer_size / sizeof(short);
		return codec_read_frame_8bps ( codec , buffer , size , offset , to_read_samples );
	}

    return 0;
}

int codec_compress_frame ( speex_encoder * codec , short* frame , uint8 * out )
{
    uint32 bytes;

    speex_bits_reset ( &codec->bits );

    if ( codec->format.nChannels == 2 )
    {
		int frame_size;

		speex_encoder_ctl ( codec->encoder, SPEEX_GET_FRAME_SIZE, &frame_size );

		speex_encode_stereo_int ( frame, frame_size , &codec->bits );
    }

    speex_encode_int( codec->encoder , frame, &codec->bits);

    speex_bits_insert_terminator(&codec->bits);

    bytes = speex_bits_nbytes( &codec->bits );

    bytes = speex_bits_write ( &codec->bits, (char*)out + 2 ,bytes );

    *((short*)out) = bytes & 0xFFFF;

    return bytes + 2;
}

uint32 codec_compress_buffer ( speex_encoder * codec , char * buffer , uint32 size , uint8 * out )
{
    uint32 offset = 0, compressed_size = 0;

    if ( (codec == NULL) || (buffer == 0) || (size == 0) )
        return 0;

    while ( codec_read_frame ( codec , buffer , size , &offset ) )
    {
		compressed_size += codec_compress_frame ( codec , codec->data_buffer , out + compressed_size );
		codec->data_buffer_size = 0; /* next frame */
    }

    /* codec->data_buffer contains size & frame_size bytes */

    return compressed_size;
}



/*-----------------------mp3 decoder --------------------------------------------*/

int mp3codec_decode (mp3decoder * _decoder, unsigned char * _mp3 , unsigned int _size, unsigned short * _out );

typedef struct 
{
    unsigned char * data_buffer;
    unsigned int start;
    unsigned int size;
} 
mp3_data_buffer;

/*static mp3_data_buffer mp3_buffer;*/

int mp3codec_initialize()
{
    return mpg123_init();    
}

int mp3codec_uninitialize()
{
    mpg123_exit();
    return 0;
}

mp3decoder * mp3codec_create_decoder ( int rate, int channels )
{
    mp3decoder * ret = (mp3decoder *) malloc( sizeof ( mp3decoder ) );
    if (ret == NULL)
        return NULL;
        
    ret->decoder = mpg123_new( NULL , NULL );

    if (ret->decoder == NULL )
    {
        free( ret );
        return NULL;
    }

    mpg123_format_none ( ret->decoder );
    mpg123_format(ret->decoder , rate , 2 , MPG123_ENC_SIGNED_16 );
    mpg123_param( ret->decoder , MPG123_VERBOSE, 2, 0);

    mpg123_open_feed( ret->decoder );

    ret->decode = &mp3codec_decode;

    return ret;
}

void mp3codec_destroy_decoder ( mp3decoder * _decoder )
{
    if ( _decoder != NULL )
    {
        if ( _decoder->decoder != NULL )
        {
            mpg123_delete( _decoder->decoder );            
        }
        free ( _decoder );
    }
}

int mp3codec_decode ( mp3decoder * _decoder, unsigned char * _mp3 , unsigned int _size, unsigned short * _out )
{
    size_t decoded = 0;
    int ret = 0;
    
    ret = mpg123_decode ( _decoder->decoder , _mp3 , _size , (unsigned char*)_out , 5*1024*1024 , &decoded );

    return decoded / sizeof( short );
}


