
/******************************************************************************
 *
 *  Quest Software, Inc.
 * ALL RIGHTS RESERVED.
 *
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 *  *  Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer.  
 * 
 *  *  Redistributions in binary form must reproduce the above copyright notice, 
 *     this list of conditions and the following disclaimer in the documentation 
 *     and/or other materials provided with the distribution.  
 * 
 *  *  Neither the name of Quest Software, Inc. nor the names of its contributors 
 *     may be used to endorse or promote products derived from this software 
 *     without specific prior written permission. 
 * 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ON AN "AS IS" 
 * AND UNSUPPORTED BASIS, WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES OF ANY KIND, 
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, NON-
 * INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL THE COPYRIGHT 
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT 
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT 
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 ************************************************************************************/


#include <errno.h>
#include <sys/wait.h>
#include <boost/bind.hpp>
#include "../common/common.h"
#include "pnserver.h"
#include "pnsession.h"

// CPP includes
#include <string>
#include <map>

#define GUI_PORT        42331
#define OWN_PORT        42332
#define MULTIADDR_PORT  42333

using asio::ip::tcp;
using asio::ip::udp;

//---------------------------------------------------------------------------
pnserver::pnserver( asio::io_service& io_service, unsigned short port )
        : io_service_( io_service ),
        acceptor_( io_service )
{
    tcp::endpoint endpoint( tcp::v4(), port );
    acceptor_.open( endpoint.protocol() );
    asio::socket_base::reuse_address option( true );
    acceptor_.set_option( option );
    acceptor_.bind( endpoint );
    acceptor_.listen();

    LOG_BANNER( "pncm" );

    session_count       = 0;
    pnsession* session  = new pnsession( io_service_, this );
    master_session      = session;

    acceptor_.async_accept( session->socket(),
                            boost::bind( &pnserver::handle_accept, this, session,
                                         asio::placeholders::error ) );
    // ping

    char buf[256];
    sprintf ( buf, "%d", OWN_PORT );

    syslog ( LOG_NOTICE, "ping" );
    udp::socket bcast_socket( io_service_, udp::v4() );
    udp::endpoint destination( asio::ip::address::from_string( "225.1.1.1" ), MULTIADDR_PORT );
    bcast_socket.send_to( asio::buffer( buf ), destination );
    bcast_socket.close();



}

//---------------------------------------------------------------------------
pnserver::~pnserver()
{
    std::map <std::string, std::string>::iterator it;
    for ( int i = 0; i < 3; i++ )
    {
        for ( it = app_lists_[i].begin(); it != app_lists_[i].end(); ++it )
        {
            app_lists_[i].erase ( it );
        }
    }

    std::map <std::string, pnsession *>::iterator i;
    std::string name;
    for ( i = sessions_.begin(); i != sessions_.end(); ++i )
    {
        name = i->first;
        delete i->second;
        sessions_.erase ( i );
        syslog ( LOG_NOTICE, "session %s closed", name.c_str() );
    }

}

//---------------------------------------------------------------------------
void pnserver::use_proxy( char *addr, char *proxy )
{
    syslog ( LOG_DEBUG, "called use_proxy(): addr %s, proxy %s", addr, proxy );
    int fd = -1;
    int i = 0, j = 0, num_entries = 0, found = -1;
    struct flock fl =
    {
        F_WRLCK, SEEK_SET, 0, 0, 0
    };
    off_t fsize;
    char *fstr = NULL, *p = NULL, *header = NULL, newroute[1024];
    char *routes[1024];

    fl.l_pid = getpid();

    if ( ( fd = open( "/etc/dante.conf", O_RDWR | O_CREAT ) ) == -1 )
    {
        syslog ( LOG_DEBUG, "use_proxy() - cannot open dante.conf: %s", strerror( errno ) );
        goto done;
    }

    if ( fcntl( fd, F_SETLKW, &fl ) == -1 )
    {
        syslog ( LOG_DEBUG, "use_proxy() - cannot lock dante.conf: %s", strerror( errno ) );
        goto done;
    }
    {
        // read the file
        fsize = lseek( fd, 0, SEEK_END );
        fstr = ( char * ) malloc ( fsize + 1 );
        lseek ( fd, 0, SEEK_SET );
        int res = read( fd, fstr, fsize );
        syslog ( LOG_DEBUG, "use_proxy() - read %d bytes", res );
        if ( res == 0 ) goto newentry;
        fstr[res] = '\0';
    }
    // search for the addr

    // split the string
    p = fstr;
    header = fstr;
    while ( ( p = strstr( p, "route" ) ) )
    {
        routes[i++] = p+5;
        *p = '\0';
        ++p;
    }
    num_entries = i;

    syslog ( LOG_DEBUG, "use_proxy() - header: %s", header );
    // pass over routing entries
    while ( i )
    {
        --i;
        syslog ( LOG_DEBUG, "use_proxy() - check route: %s", routes[i] );
        // now routes[i] points to a symbol next to the "route" word
        // every error means that file is corrupted,
        // and we will just try to search in other routes
        if ( ( p = strstr ( routes[i], "to:" ) ) == NULL )
            continue;
        p += 3;
        while ( isspace( *p ) ) ++p;
        if ( strncmp( p, addr, strlen( addr ) ) == 0 )
        {
            syslog ( LOG_DEBUG, "use_proxy() - found" );
            found = i;
            break;
        }
    }

    // write back everything to the file
    lseek ( fd, 0, SEEK_SET );
    ftruncate( fd, 0 );
    if ( write( fd, header, strlen ( header ) ) < 0 )
    {
        syslog ( LOG_DEBUG, "pnserver::use_proxy() - write error: %s", strerror( errno ) );
    }
    for ( j = 0; j < num_entries; ++j )
    {
        if ( j == found ) continue;
        *( routes[j]-5 ) = 'r';
        write( fd, routes[j]-5, strlen ( routes[j]-5 ) );
    }

newentry:

    // write a new route
    p = strchr( proxy, ':' );
    if ( p ) *p = '\0';
    i = sprintf( newroute, "route { from: 0.0.0.0/0 to: %s/24 via: %s port = %s method: none }\n", addr, proxy, p ? p+1 : "1080" );
    newroute[i] = '\0';
    syslog( LOG_DEBUG, "new route: %s", newroute );
    if ( p ) *p = ':';
    write ( fd, newroute, i );

done:

    if ( fstr )
        free ( fstr );

    fl.l_type = F_UNLCK;

    fcntl( fd, F_SETLK, &fl );

    close( fd );
}

//---------------------------------------------------------------------------
void pnserver::inc_session_count( void )
{
    session_count ++;
    syslog ( LOG_NOTICE, "Number session instances now %i.", session_count );
}

void pnserver::dec_session_count( void )
{
    session_count --;
    syslog ( LOG_NOTICE, "Number session instances now %i.", session_count );

    // If I ever delete the last one, shut the server down
    if ( session_count == 0 )
    {
        syslog ( LOG_NOTICE, "There are no other sessions." );
        throw std::logic_error ( "Stop request" );
    }
}


//---------------------------------------------------------------------------
void pnserver::handle_accept( pnsession* session, const asio::error_code& error )
{
    if ( !error )
    {
        syslog ( LOG_NOTICE, "accept" );

        session->start();
        session = new pnsession( io_service_, this );
        acceptor_.async_accept( session->socket(),
                                boost::bind( &pnserver::handle_accept, this, session,
                                             asio::placeholders::error ) );
    }
    else
    {
        syslog ( LOG_NOTICE, "Error in accept(): %s", error.message().c_str() );
        delete session;
    }
}

//---------------------------------------------------------------------------
std::string pnserver::get_list ( const std::string& broker, int type )
{
    std::map <std::string, std::string>::iterator i = app_lists_[type].find ( broker );
    if ( i == app_lists_[type].end() )
    {
        return std::string ( "" );
    }

    return i->second;
}

//---------------------------------------------------------------------------
void pnserver::set_list ( const std::string& broker, const std::string& list, int type )
{
    syslog ( LOG_DEBUG, "broker: %s   apps: %s", broker.c_str(), list.c_str() );
    app_lists_[type][broker] = list;
}

//---------------------------------------------------------------------------
void pnserver::register_session ( const std::string& session_name, pnsession * session )
{
    syslog ( LOG_NOTICE, "Register session named: %s reference: %p", session_name.c_str(), session );
    sessions_.insert( std::pair <std::string, pnsession *> ( session_name, session ) );

    // Some debugging here
    syslog ( LOG_NOTICE, "Session list has %i item(s) now.", sessions_.size() );
}

//---------------------------------------------------------------------------
pnsession * pnserver::find_session ( const std::string& session_name )
{
    syslog ( LOG_NOTICE, "Searching for session %s", session_name.c_str() );
    std::map <std::string, pnsession *>::iterator i = sessions_.find ( session_name );
    if ( i == sessions_.end() )
    {
        syslog ( LOG_NOTICE, "Session '%s' not found", session_name.c_str() );
        return NULL;
    }
    syslog ( LOG_NOTICE, "Session '%s' found", session_name.c_str() );
    return i->second;
}

//---------------------------------------------------------------------------
void pnserver::remove_session ( pnsession * session )
{

    std::map <std::string, pnsession *>::iterator i;
    std::string name;

    syslog( LOG_NOTICE, "Attempting to remove session %p from list.", session );

    for ( i = sessions_.begin(); i != sessions_.end(); ++i )
    {
        name = i->first;
        syslog ( LOG_NOTICE, "Checking session: '%s' (%p)", name.c_str(), i->second );
        if ( i->second == session )
        {
            name = i->first;
            sessions_.erase ( i );
            syslog ( LOG_NOTICE, "Found match, session %s closed", name.c_str() );

            // Some debugging here
            syslog ( LOG_NOTICE, "Session list has %i item(s) now.", sessions_.size() );

            // If I deleted a session AND there are none left, then shut down
            if ( ( sessions_.size() == 0 ) && ( gui_sessions_.size() == 0 ) )
            {
                syslog ( LOG_NOTICE, "There are no other sessions." );
                throw std::logic_error ( "Stop request" );
            }

            break;
        }
    }

}




//---------------------------------------------------------------------------
void sigchld_handler ( int signo )
{
    pid_t pid;
    int stat;

    while ( ( pid = waitpid ( -1, &stat, WNOHANG ) ) > 0 );
    return;
}

//---------------------------------------------------------------------------
int main ()
{
    int i;
    struct sigaction sa;
    asio::io_service io_service;

    openlog ( "pnserver", LOG_PID, LOG_USER );

    // daemonize

    if ( ( i = fork() ) < 0 )
    {
        syslog ( LOG_NOTICE, "error: %s", strerror ( errno ) );
        exit ( 1 );
    }
    else if ( i )
    {
        exit ( 0 );
    }
    setsid();
    close ( 0 );
    close ( 1 );
    close ( 2 );
    i = open ( "/dev/null", O_RDWR );
    dup ( i );
    dup ( i );
    chdir ( "/" );
    sigemptyset ( &sa.sa_mask );
    sa.sa_handler = SIG_IGN;
    sa.sa_flags = 0;
    sigaction ( SIGTSTP, &sa, NULL );
    sigaction ( SIGTTIN, &sa, NULL );
    sigaction ( SIGTTOU, &sa, NULL );
    sigaction ( SIGHUP, &sa, NULL );
    sa.sa_handler = sigchld_handler;
    sa.sa_flags |= SA_RESTART;
    sigaction ( SIGCHLD, &sa, NULL );

    // start server

    pnserver srvr ( io_service, OWN_PORT );
    try
    {
        io_service.run();
    }
    catch ( std::exception& e )
    {
        syslog ( LOG_NOTICE, "Received an internal message: %s", e.what() );
    }

    syslog ( LOG_NOTICE, "Stopping Session Server" );
    closelog ();

    return 0;
}
